# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from itertools import combinations
from typing import List, Dict, Optional, Tuple

import numpy as np
from monty.json import MSONable
from monty.serialization import loadfn
from pydefect.util.prepare_names import prettify_names
from scipy.spatial import HalfspaceIntersection
from tabulate import tabulate
from vise.util.mix_in import ToJsonFileMixIn, ToYamlFileMixIn
from vise.util.string import latexify, numbers_to_lowercases


@dataclass
class DefectEnergy(MSONable):
    formation_energy: float
    energy_corrections: Dict[str, float]
    is_shallow: Optional[bool] = None

    @property
    def total_correction(self) -> float:
        return sum([v for v in self.energy_corrections.values()])

    def energy(self, with_correction):
        if with_correction:
            return self.formation_energy + self.total_correction
        return self.formation_energy


@dataclass
class DefectEnergyInfo(MSONable, ToYamlFileMixIn):
    name: str
    charge: int
    atom_io: Dict[str, int]
    defect_energy: DefectEnergy
    # This defect formation energy is estimated at the Fermi level locating at
    # the vbm and references are set to their standard states.

    def to_yaml(self) -> str:
        lines = [f"name: {self.name}",
                 f"charge: {self.charge}",
                 f"formation_energy: {self.defect_energy.formation_energy}",
                 f"atom_io:"]
        for k, v in self.atom_io.items():
            lines.append(f"  {k}: {v}")
        lines.append(f"energy_corrections:")
        for k, v in self.defect_energy.energy_corrections.items():
            lines.append(f"  {k}: {v}")
        is_shallow = "" if self.defect_energy.is_shallow is None \
            else self.defect_energy.is_shallow
        lines.append(f"is_shallow: {is_shallow}")
        return "\n".join(lines)

    @classmethod
    def from_yaml(cls, filename: str = "defect_energy_info.yaml"
                  ) -> "DefectEnergyInfo":
        d = loadfn(filename)
        if d["atom_io"] is None:
            d["atom_io"] = {}
        return cls(d.pop("name"), d.pop("charge"), d.pop("atom_io"),
                   DefectEnergy(**d))


@dataclass
class DefectEnergies(MSONable):
    atom_io: Dict[str, int]
    charges: List[int]
    defect_energies: List[DefectEnergy]

    def to_yaml(self):
        pass


@dataclass
class DefectEnergySummary(MSONable, ToJsonFileMixIn):
    title: str
    defect_energies: Dict[str, "DefectEnergies"]
    rel_chem_pots: Dict[str, Dict[str, float]]
    cbm: float
    supercell_vbm: float
    supercell_cbm: float
    """ The base Fermi level is set at the VBM."""

    def __str__(self):
        lines = [f"title: {numbers_to_lowercases(self.title)}",
                 "rel_chem_pots:"]
        chem_pot = []
        for k, v in self.rel_chem_pots.items():
            elem_list = [f"{elem}: {val:.2f}" for elem, val in v.items()]
            chem_pot.append(f" -{k} " + " ".join(elem_list))
        lines.append('\n'.join(chem_pot))
        lines.append(f"vbm: 0.00, cbm: {self.cbm:.2f}, "
                     f"supercell vbm: {self.supercell_vbm:.2f}, "
                     f"supercell_cbm: {self.supercell_cbm:.2f}")
        lines.append("")

        defects = []
        for name, des in self.defect_energies.items():
            atom_io = " ".join(
                [f"{elem}: {io}" for elem, io in des.atom_io.items()])
            charges_energies = list(zip(des.charges, des.defect_energies))
            charges_energies.sort(key=lambda x: x[0])
            for charge, de in charges_energies:
                defects.append([name, atom_io, charge, de.formation_energy,
                                de.total_correction, de.is_shallow])
                name, atom_io = "", ""
        headers = ("name", "atom_io", "charge", "energy", "correction",
                   "is_shallow")
        floatfmt = ("", "", "", ".3f", ".3f", "")
        lines.append(tabulate(defects, headers=headers, floatfmt=floatfmt))

        return "\n".join(lines)

    # TODO: when making this class, if all the defect charges show shallow
    # show warning.
    def charge_energies(self,
                        chem_pot_label: str,
                        allow_shallow: bool,
                        with_corrections: bool,
                        e_range: Tuple[float, float],
                        name_style: Optional[str] = None,
                        ) -> "ChargeEnergies":
        #TODO: generate logger.info when e_min < supercell_vbm.
        rel_chem_pot = self.rel_chem_pots[chem_pot_label]
        result = {}
        for k, v in self.defect_energies.items():
            x = []
            for c, e in zip(v.charges, v.defect_energies):
                if allow_shallow is False and e.is_shallow is True:
                    continue
                reservoir_e = sum([-diff * rel_chem_pot[elem]
                                  for elem, diff in v.atom_io.items()])
                x.append((c, e.energy(with_corrections) + reservoir_e))

            if x:
                result[k] = SingleChargeEnergies(x)

        return ChargeEnergies(prettify_names(result, name_style), e_range[0], e_range[1])

    @property
    def latexified_title(self):
        return latexify(self.title)


@dataclass
class ChargeEnergies:
    charge_energies_dict: Dict[str, "SingleChargeEnergies"]
    e_min: float
    e_max: float

    # change to lazy evaluation?
    def __post_init__(self):
        self.cross_point_dicts = {}
        large_minus_number = -1e4
        for name, ce in self.charge_energies_dict.items():
            half_spaces = []
            for charge, corr_energy in ce.charge_energies:
                half_spaces.append([-charge, 1, -corr_energy])

            half_spaces.append([-1, 0, self.e_min])
            half_spaces.append([1, 0, -self.e_max])
            half_spaces.append([0, -1, large_minus_number])

            feasible_point = np.array([(self.e_min + self.e_max) / 2, -1e3])

            hs = HalfspaceIntersection(np.array(half_spaces), feasible_point)
            boundary_points = []
            inner_cross_points = []
            for intersection in hs.intersections:
                x, y = np.round(intersection, 8)
                if self.e_min + 0.001 < x < self.e_max - 0.001:
                    inner_cross_points.append([x, y])
                elif y > large_minus_number + 1:
                    boundary_points.append([x, y])

            self.cross_point_dicts[name] = CrossPoints(inner_cross_points,
                                                       boundary_points)

    def energy_range(self, space: float) -> List[float]:
        candidates = []
        for cp in self.cross_point_dicts.values():
            candidates.extend(cp.t_all_sorted_points[1])
        return [min(candidates) - space, max(candidates) + space]


@dataclass
class SingleChargeEnergies:
    charge_energies: List[Tuple[int, float]]

    @property
    def transition_levels(self) -> Dict[Tuple[int, int], float]:
        result = {}
        for (c1, e1), (c2, e2) in combinations(self.charge_energies, 2):
            result[(c1, c2)] = - (e1 - e2) / (c1 - c2)
        return result

    def pinning_level(self, e_min, e_max
                      ) -> Tuple[Tuple[float, Optional[int]],
                                 Tuple[float, Optional[int]]]:
        """
        :return: ((Lower pinning, its charge), (Upper pinning, its charge))
        """
        lower_pinning, upper_pinning = float("-inf"), float("inf")
        lower_charge, upper_charge = None, None
        for charge, energy in self.charge_energies:
            if charge == 0:
                continue
            pinning = - energy / charge
            if charge > 0 and pinning > lower_pinning:
                lower_pinning, lower_charge = pinning, charge
            elif pinning < upper_pinning:
                upper_pinning, upper_charge = pinning, charge

        if lower_charge is None or lower_pinning < e_min:
            lower = None
        else:
            lower = (lower_pinning, lower_charge)

        if upper_charge is None or upper_pinning > e_max:
            upper = None
        else:
            upper = (lower_pinning, lower_charge)
        return lower, upper

    def energy_at_ef(self, ef: float) -> Tuple[float, int]:
        """
        :return: (Lowest energy, its charge)
        """
        result_e, result_charge = float("inf"), None
        for charge, energy in self.charge_energies:
            energy = energy + charge * ef
            if energy < result_e:
                result_e, result_charge = energy, charge
        return result_e, result_charge


@dataclass
class CrossPoints:
    inner_cross_points: List[List[float]]  # [Fermi level, energy]
    boundary_points: List[List[float]]

    @property
    def all_sorted_points(self):
        return sorted(self.boundary_points + self.inner_cross_points,
                      key=lambda v: v[0])

    @property
    def t_all_sorted_points(self):
        return np.transpose(np.array(self.all_sorted_points)).tolist()

    @property
    def t_inner_cross_points(self):
        return np.transpose(np.array(self.inner_cross_points)).tolist()

    @property
    def t_boundary_points(self):
        return np.transpose(np.array(self.boundary_points)).tolist()

    @property
    def charges(self) -> List[int]:
        result = []
        for i, j in zip(self.all_sorted_points[:-1], self.all_sorted_points[1:]):
            dx = j[0] - i[0]
            dy = j[1] - i[1]
            result.append(int(round(dy / dx)))
        return result

    @property
    def charge_list(self):
        charges = [None] + self.charges + [None]
        return list(zip(charges[:-1], charges[1:]))

    @property
    def annotated_charge_positions(self):
        result = {}
        for ((x1, y1), (x2, y2)), charge \
                in zip(zip(self.all_sorted_points[:-1],
                           self.all_sorted_points[1:]),
                       self.charges):
            result[charge] = [(x1 + x2) / 2, (y1 + y2) / 2]
        return result

    def __str__(self):
        lines = []
        for point in self.all_sorted_points:
            lines.append(f"{point[0]:12.4f} {point[1]:12.4f}")
        return "\n".join(lines)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.


import argparse
import sys
import warnings

from monty.serialization import loadfn
from pydefect.analyzer.grids import Grids
from pydefect.cli.main import epilog, description, add_sub_parser
from pydefect.cli.vasp.main_vasp_util_functions import \
    make_parchg_dir, make_refine_defect_poscar, \
    calc_charge_state, make_defect_entry_main, calc_grids, \
    make_defect_charge_info_main
from pymatgen.core import Structure
from pymatgen.io.vasp import Chgcar
from pymatgen.io.vasp.inputs import UnknownPotcarWarning

warnings.simplefilter('ignore', UnknownPotcarWarning)


def parse_args_main_vasp_util(args):
    parser = argparse.ArgumentParser(epilog=epilog,
                                     description=description + """      
    This command provides some utilities related to the VASP calculations""")

    subparsers = parser.add_subparsers()
    dir_parser = add_sub_parser(argparse, name="dir")

    # -- calc charge state -----------------------------------------------------
    parser_calc_charge_state = subparsers.add_parser(
        name="calc_charge_state",
        description="Calc defect charge states from INCAR, POSCAR and POTCAR"
                    "files.",
        parents=[dir_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['ccs'])

    parser_calc_charge_state.set_defaults(func=calc_charge_state)

    # -- make defect entry -----------------------------------------------------
    parser_make_defect_entry = subparsers.add_parser(
        name="make_defect_entry",
        description="Make defect entry from INCAR, POSCAR and POTCAR"
                    "files.",
        parents=[dir_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['de'])
    parser_make_defect_entry.add_argument(
        "-n", "--name", type=str, required=True,
        help="Name used for plotting energy diagram.")
    parser_make_defect_entry.add_argument(
        "-p", "--perfect", type=Structure.from_file, required=True,
        help="Perfect supercell POSCAR file name.")

    parser_make_defect_entry.set_defaults(func=make_defect_entry_main)

    # -- make parchg dir -----------------------------------------------
    parser_make_parchg_dir = subparsers.add_parser(
        name="parchg_dir",
        description="Make parchg directory containing a set of VASP files "
                    "required for generating PARCHG files.",
        parents=[dir_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pd'])

    parser_make_parchg_dir.set_defaults(func=make_parchg_dir)

    # -- make refine defect poscar ---------------------------------------------
    parser_make_refine_defect_poscar = subparsers.add_parser(
        name="refine_defect_poscar",
        description="Refine defect structure POSCAR.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['rdp'])

    parser_make_refine_defect_poscar.add_argument(
        "-p", "--poscar", dest="structure", type=Structure.from_file)
    parser_make_refine_defect_poscar.add_argument(
        "-d", "--defect_entry", type=loadfn)
    parser_make_refine_defect_poscar.add_argument(
        "-n", "--poscar_name", type=str)
    parser_make_refine_defect_poscar.set_defaults(
        func=make_refine_defect_poscar)

    # -- calc grids ------------------------------------------------------------
    parser_calc_grids = subparsers.add_parser(
        name="calc_grids",
        description="Calc Grids.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['cg'])

    parser_calc_grids.add_argument(
        "-c", "--chgcar", type=Chgcar.from_file, required=True)
    parser_calc_grids.set_defaults(func=calc_grids)

    # -- calc defect charge info -----------------------------------------------
    parser_calc_def_charge_info = subparsers.add_parser(
        name="calc_defect_charge_info",
        description="Calc defect charge info.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['cdc'])

    parser_calc_def_charge_info.add_argument(
        "-p", "--parchgs", type=str, nargs="+", required=True,
        help="PARCHG files.")
    parser_calc_def_charge_info.add_argument(
        "-b", "--bin_interval", type=float, default=0.2)
    parser_calc_def_charge_info.add_argument(
        "-g", "--grids", type=Grids.from_file, required=True)

    parser_calc_def_charge_info.set_defaults(func=make_defect_charge_info_main)

    return parser.parse_args(args)


def main():
    args = parse_args_main_vasp_util(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from typing import List, Union

from vise.util.typing import Coords


def pretty_coords(coords: Union[List[float], Coords]):
    return f'({", ".join([f"{c:6.3f}" for c in coords])})'



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from itertools import product
from typing import Dict, List

from pydefect.database.database import oxidation_state, electronegativity
from pydefect.defaults import defaults
from pydefect.input_maker.defect import SimpleDefect
from pydefect.input_maker.defect_set import DefectSet, screen_defect_set
from pydefect.input_maker.supercell_info import SupercellInfo


class DefectSetMaker:
    def __init__(self,
                 supercell_info: SupercellInfo,
                 overwritten_oxi_states: Dict[str, int] = None,
                 dopants: List[str] = None,
                 ele_neg_diff: float = defaults.ele_neg_diff,
                 keywords: List[str] = None):

        self.supercell_info = supercell_info
        self._overwritten_oxi_states = overwritten_oxi_states or {}
        self._dopants = dopants or []
        self._ele_neg_diff = ele_neg_diff

        self._subs = self._dopants + self._host_elements

        defect_set = self._create_defect_set()
        if keywords:
            self.defect_set = screen_defect_set(defect_set, keywords)
        else:
            self.defect_set = defect_set

    def _oxidation_state(self, element):
        try:
            return self._overwritten_oxi_states[element]
        except KeyError:
            return oxidation_state(element)

    @property
    def _host_elements(self) -> List[str]:
        return [str(specie) for specie in
                set(self.supercell_info.structure.species)]

    def _create_defect_set(self):
        result = self._create_vacancy_set()
        result.update(self._create_substitutional_set())
        result.update(self._create_interstitial_set())
        return DefectSet(result)

    def _create_vacancy_set(self):
        result = set()
        for out_name, site in self.supercell_info.sites.items():
            minus_oxi_stat = - self._oxidation_state(site.element)
            result.add(SimpleDefect(None, out_name, charge_set(minus_oxi_stat)))

        return result

    def _create_substitutional_set(self):
        result = set()
        for in_name, (out_name, site) \
                in product(self._subs, self.supercell_info.sites.items()):
            host_element = site.element
            sub_en = electronegativity(in_name)
            host_en = electronegativity(host_element)
            are_same_elem = in_name == host_element
            if None in (sub_en, host_en) or are_same_elem or \
                    abs(sub_en - host_en) > self._ele_neg_diff:
                continue

            sub_oxi = self._oxidation_state(in_name)
            host_oxi = self._oxidation_state(host_element)
            oxi_diff = sub_oxi - host_oxi
            result.add(SimpleDefect(in_name, out_name, charge_set(oxi_diff)))

        return result

    def _create_interstitial_set(self):
        result = set()
        for in_name in self._subs:
            oxi_stat = self._oxidation_state(in_name)
            for i in range(len(self.supercell_info.interstitials)):
                result.add(SimpleDefect(in_name, f"i{i + 1}",
                                        charge_set(oxi_stat)))

        return result


def charge_set(ox_state: int) -> List[int]:
    """Set of defect charge states.

    -1 (1) is included for positive (negative) odd number.
    E.g., charge_set(3) = [-1, 0, 1, 2, 3]
          charge_set(-3) = [-3, -2, -1, 0, 1]
          charge_set(2) = [0, 1, 2]
          charge_set(-4) = [-4, -3, -2, -1, 0]

    Args:
        ox_state (int): an integer

    Returns:
        Set of candidate charges
    """
    if ox_state >= 0:
        charges = [i for i in range(ox_state + 1)]
        if ox_state % 2 == 1:
            charges.insert(0, -1)
    else:
        charges = [i for i in range(ox_state, 1)]
        if ox_state % 2 == 1:
            charges.append(1)

    return charges



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

from pydefect.analyzer.band_edge_states import PerfectBandEdgeState
from pydefect.cli.vasp.main_vasp import parse_args_main_vasp
from pydefect.defaults import defaults


def test_unitcell(mocker):
    mock = mocker.patch("pydefect.cli.vasp.main_vasp.Vasprun")
    mock_outcar = mocker.patch("pydefect.cli.vasp.main_vasp.Outcar")

    parsed_args = parse_args_main_vasp(["u",
                                        "-vb", "vasprun.xml",
                                        "-ob", "OUTCAR-1",
                                        "-odc", "OUTCAR-2",
                                        "-odi", "OUTCAR-3",
                                        "-n", "name"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        vasprun_band=mock.return_value,
        outcar_band=mock_outcar.return_value,
        outcar_dielectric_clamped=mock_outcar.return_value,
        outcar_dielectric_ionic=mock_outcar.return_value,
        name="name",
        func=parsed_args.func,
    )
    assert parsed_args == expected
    mock.assert_called_once_with("vasprun.xml")
    mock_outcar.assert_any_call("OUTCAR-1")
    mock_outcar.assert_any_call("OUTCAR-2")
    mock_outcar.assert_called_with("OUTCAR-3")


def test_make_poscars_wo_options():
    parsed_args = parse_args_main_vasp(["mp", "-e", "Mg", "O"])
    expected = Namespace(
        elements=["Mg", "O"],
        e_above_hull=defaults.e_above_hull,
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_poscars_w_options():
    parsed_args = parse_args_main_vasp([
        "mp", "-e", "Mg", "O", "--e_above_hull", "0.1"])
    expected = Namespace(
        elements=["Mg", "O"],
        e_above_hull=0.1,
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_local_extrema_wo_options(mocker):
    mock_chgcar = mocker.patch("pydefect.cli.vasp.main_vasp.Chgcar")
    parsed_args = parse_args_main_vasp(["le", "-v", "CHGCAR"])
    expected = Namespace(
        volumetric_data=[mock_chgcar.from_file.return_value],
        find_max=False,
        info=None,
        threshold_frac=None,
        threshold_abs=None,
        min_dist=0.5,
        tol=0.5,
        radius=0.4,
        func=parsed_args.func)
    assert parsed_args == expected
    mock_chgcar.from_file.assert_called_once_with("CHGCAR")


def test_make_local_extrema_w_options(mocker):
    mock_chgcar = mocker.patch("pydefect.cli.vasp.main_vasp.Chgcar")
    parsed_args = parse_args_main_vasp(["le",
                                        "-v", "CHGCAR",
                                        "--find_max",
                                        "--info", "a",
                                        "--threshold_frac", "0.1",
                                        "--threshold_abs", "0.2",
                                        "--min_dist", "0.3",
                                        "--tol", "0.4",
                                        "--radius", "0.5"])
    expected = Namespace(
        volumetric_data=[mock_chgcar.from_file.return_value],
        find_max=True,
        info="a",
        threshold_frac=0.1,
        threshold_abs=0.2,
        min_dist=0.3,
        tol=0.4,
        radius=0.5,
        func=parsed_args.func)
    assert parsed_args == expected


def test_defect_entries():
    parsed_args = parse_args_main_vasp(["de"])
    expected = Namespace(func=parsed_args.func)
    assert parsed_args == expected


def test_calc_results():
    parsed_args = parse_args_main_vasp(["cr", "-d", "Va_O1_0", "Va_O1_1"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        verbose=False,
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_perfect_band_edge_state():
    parsed_args = parse_args_main_vasp(["pbes", "-d", "Va_O1_0"])
    expected = Namespace(
        dir=Path("Va_O1_0"),
        func=parsed_args.func)
    assert parsed_args == expected


def test_band_edge_orbital_infos(mocker):
    mock_p_edge_state = mocker.Mock(spec=PerfectBandEdgeState, autospec=True)

    def side_effect(filename):
        if filename == "perfect_band_edge_state.json":
            return mock_p_edge_state
        else:
            raise ValueError

    mocker.patch("pydefect.cli.main.loadfn", side_effect=side_effect)
    parsed_args = parse_args_main_vasp([
        "beoi",
        "-d", "Va_O1_0", "Va_O1_1",
        "-pbes", "perfect_band_edge_state.json",
        "-y", "0.0", "1.0",
        "-v"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        p_state=mock_p_edge_state,
        y_range=[0.0, 1.0],
        verbose=True,
        func=parsed_args.func)
    assert parsed_args == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pydefect.corrections.abstract_correction import Correction

from pydefect.corrections.efnv_correction import \
    PotentialSite
from vise.tests.helpers.assertion import assert_msonable


def test_extended_fnv_correction_msonable(efnv_correction):
    assert isinstance(efnv_correction, Correction)


def test_repr(efnv_correction):
    expected = """+-------------------+-----+
| charge            |  10 |
| pc term           |   1 |
| alignment term    | -15 |
| correction energy | -14 |
+-------------------+-----+"""
    assert str(efnv_correction) == expected


def test_extended_fnv_correction_ave_pot_diff(efnv_correction):
    expected = ((1.5 - 0.2) + (2.0 - 0.3)) / 2
    assert efnv_correction.average_potential_diff == expected


def test_extended_fnv_correction_alignment_correction(efnv_correction):
    expected = -10 * ((1.5 - 0.2) + (2.0 - 0.3)) / 2
    assert efnv_correction.alignment_correction == expected


def test_extended_fnv_correction_correction_energy(efnv_correction):
    expected = 1.0 - 10 * ((1.5 - 0.2) + (2.0 - 0.3)) / 2
    assert efnv_correction.correction_energy == expected


def test_extended_fnv_correction_correction_dict(efnv_correction):
    expected = {"pc term": 1.0, "alignment term": -15.0}
    assert efnv_correction.correction_dict == expected


def test_defect_site_diff_pot():
    s = PotentialSite("H", distance=1.999, potential=1.0, pc_potential=0.1)
    assert s.diff_pot == 1.0 - 0.1


def test_defect_site_msonable(efnv_correction):
    assert_msonable(PotentialSite(specie="H", distance=1.999, potential=1.0,
                                  pc_potential=0.1))




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

import numpy as np
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.defect_structure_info import DefectStructureInfo
from pydefect.analyzer.unitcell import Unitcell
from pydefect.cli.vasp.main_vasp_functions import make_unitcell, \
    make_competing_phase_dirs, make_composition_energies, make_defect_entries, \
    make_calc_results, make_band_edge_orb_infos_and_eigval_plot, \
    make_perfect_band_edge_state, make_local_extrema
from pydefect.input_maker.defect import SimpleDefect
from pydefect.input_maker.defect_entry import DefectEntry
from pydefect.input_maker.defect_set import DefectSet
from pymatgen.core import Composition, Structure
from pymatgen.io.vasp import Vasprun, Outcar, VolumetricData
from vise.defaults import defaults


def test_make_unitcell(tmpdir, mocker):
    tmpdir.chdir()
    vasprun_band_mock = mocker.Mock(spec=Vasprun, autospec=True)
    outcar_band_mock = mocker.Mock(spec=Outcar, autospec=True)
    outcar_dielectric_mock = mocker.Mock(spec=Outcar, autospec=True)
    args = Namespace(vasprun_band=vasprun_band_mock,
                     outcar_band=outcar_band_mock,
                     outcar_dielectric_clamped=outcar_dielectric_mock,
                     outcar_dielectric_ionic=outcar_dielectric_mock,
                     name="H")

    mock = mocker.patch(
        "pydefect.cli.vasp.main_vasp_functions.make_unitcell_from_vasp")
    mock.return_value = Unitcell(system="H",
                                 vbm=1.0,
                                 cbm=2.0,
                                 ele_dielectric_const=np.eye(3).tolist(),
                                 ion_dielectric_const=np.eye(3).tolist())
    make_unitcell(args)
    mock.assert_called_once_with(
        vasprun_band=vasprun_band_mock,
        outcar_band=outcar_band_mock,
        outcar_dielectric_clamped=outcar_dielectric_mock,
        outcar_dielectric_ionic=outcar_dielectric_mock,
        system_name="H")


def test_make_competing_phase_dirs(mocker):
    args = Namespace(elements=["Mg", "O"],
                     e_above_hull=0.1)
    mock = mocker.patch("pydefect.cli.vasp.main_vasp_functions.MpQuery")
    mock_make = mocker.patch(
        "pydefect.cli.vasp.main_vasp_functions.make_poscars_from_query")
    make_competing_phase_dirs(args)
    mock.assert_called_once_with(element_list=args.elements,
                                 e_above_hull=args.e_above_hull)
    mock_make.assert_called_once_with(
        materials_query=mock.return_value.materials, path=Path.cwd())


def test_make_composition_energies(mocker, tmpdir):
    def side_effect(key):
        mock_outcar = mocker.Mock()
        if key == Path("Mg") / defaults.outcar:
            mock_outcar.final_energy = -10
        elif key == Path("O") / defaults.outcar:
            mock_outcar.final_energy = -20
        else:
            raise ValueError
        return mock_outcar

    def side_effect_structure(key):
        mock_structure = mocker.Mock()
        if key == Path("Mg") / defaults.contcar:
            mock_structure.composition = Composition("Mg2")
        elif key == Path("O") / defaults.contcar:
            mock_structure.composition = Composition("O2")
        else:
            raise ValueError
        return mock_structure

    tmpdir.chdir()
    print(tmpdir)
    mock = mocker.patch("pydefect.cli.vasp.main_vasp_functions.Outcar",
                        side_effect=side_effect)
    mock_str = mocker.patch("pydefect.cli.vasp.main_vasp_functions.Structure.from_file",
                            side_effect=side_effect_structure)
    args = Namespace(yaml_file=None, dirs=[Path("Mg"), Path("O")], verbose=False)
    make_composition_energies(args)
    actual = Path("composition_energies.yaml").read_text()
    expected = """Mg2:
  energy: -10.0
  source: local
O2:
  energy: -20.0
  source: local
"""
    assert actual == expected


def test_make_local_extrema(tmpdir, mocker, simple_cubic):
    print(tmpdir)
    tmpdir.chdir()

    mock_params = mocker.patch("pydefect.cli.vasp.main_vasp_functions.VolumetricDataAnalyzeParams")
    mock_make_extrema = mocker.patch("pydefect.cli.vasp.main_vasp_functions.make_local_extrema_from_volumetric_data")
    volumetric_data = VolumetricData(simple_cubic,
                                     data={"total": np.array([[[0.0]]])})
    args = Namespace(volumetric_data=[volumetric_data],
                     find_max=True,
                     info="a",
                     threshold_frac=None,
                     threshold_abs=None,
                     min_dist=0.1,
                     tol=0.2,
                     radius=0.3)
    make_local_extrema(args)
    mock_params.assert_called_once_with(None, None, 0.1, 0.2, 0.3)
    mock_make_extrema.assert_called_once_with(volumetric_data=volumetric_data,
                                              params=mock_params.return_value,
                                              info="a",
                                              find_min=False)
    mock_make_extrema.return_value.to_json_file.assert_called_once_with()


def test_make_defect_entries(tmpdir, supercell_info):
    tmpdir.chdir()
    supercell_info.to_json_file()
    defect_set = DefectSet({SimpleDefect(None, "He1", [-1, 0])})
    defect_set.to_yaml()
    args = Namespace()
    make_defect_entries(args)
    names = {str(name) for name in Path(".").glob("*")}
    assert names == {'Va_He1_-1', 'defect_in.yaml', 'perfect', 'Va_He1_0',
                     'supercell_info.json'}

    perfect_structure = Structure.from_file(str(Path("perfect") / "POSCAR"))
    assert perfect_structure == supercell_info.structure

    file_names = {str(file_name.name)
                  for file_name in Path("Va_He1_-1").glob("*")}
    assert file_names == {"POSCAR", "defect_entry.json", "prior_info.yaml"}

    expected = """charge: -1
"""
    assert Path("Va_He1_-1/prior_info.yaml").read_text() == expected


def test_make_calc_results(tmpdir, mocker):
    tmpdir.chdir()
    mock = mocker.patch(
        "pydefect.cli.vasp.main_vasp_functions.make_calc_results_from_vasp")
    mock_vasprun = mocker.patch("pydefect.cli.vasp.main_vasp_functions.Vasprun")
    mock_outcar = mocker.patch("pydefect.cli.vasp.main_vasp_functions.Outcar")
    mock_calc_results = mocker.Mock(spec=CalcResults)
    mock.return_value = mock_calc_results
    args = Namespace(dirs=[Path("a")], verbose=False)
    make_calc_results(args)

    mock_vasprun.assert_called_with(Path("a") / defaults.vasprun)
    mock_outcar.assert_called_with(Path("a") / defaults.outcar)
    mock.assert_called_with(vasprun=mock_vasprun.return_value,
                            outcar=mock_outcar.return_value)
    mock_calc_results.to_json_file.assert_called_with("a/calc_results.json")


def test_make_perfect_band_edge_state(mocker):
    mock_vasprun = mocker.patch("pydefect.cli.vasp.main_vasp_functions.Vasprun")
    mock_procar = mocker.patch("pydefect.cli.vasp.main_vasp_functions.Procar")
    mock_outcar = mocker.patch("pydefect.cli.vasp.main_vasp_functions.Outcar")
    mock_make_perf_be_state = mocker.patch(
        "pydefect.cli.vasp.main_vasp_functions.make_perfect_band_edge_state_from_vasp")

    args = Namespace(dir=Path("perfect"))
    make_perfect_band_edge_state(args)

    mock_vasprun.assert_called_with(Path("perfect") / defaults.vasprun)
    mock_procar.assert_called_with(Path("perfect") / defaults.procar)
    mock_outcar.assert_called_with(Path("perfect") / defaults.outcar)
    mock_make_perf_be_state.assert_called_with(mock_procar.return_value,
                                               mock_vasprun.return_value,
                                               mock_outcar.return_value)


def test_make_band_edge_orb_infos_and_eigval_plot(mocker):
    mock_procar = mocker.patch("pydefect.cli.vasp.main_vasp_functions.Procar")
    mock_vasprun = mocker.patch("pydefect.cli.vasp.main_vasp_functions.Vasprun")

    mock_p_state = mocker.Mock()
    mock_p_state.vbm_info.energy = 10
    mock_p_state.cbm_info.energy = 20

    mock_defect_entry = mocker.Mock(spec=DefectEntry, autospec=True)

    mock_structure_info = mocker.Mock(spec=DefectStructureInfo, autospec=True)

    mock_make_orbital_infos = mocker.patch(
        "pydefect.cli.vasp.main_vasp_functions.make_band_edge_orbital_infos")
    mock_eigval_plotter = mocker.patch(
        "pydefect.cli.vasp.main_vasp_functions.EigenvalueMplPlotter")

    def side_effect(key):
        if str(key) == "Va_O1_2/defect_entry.json":
            mock_defect_entry.name = "Va_O1"
            mock_defect_entry.charge = 2
            return mock_defect_entry
        elif str(key) == "Va_O1_2/defect_structure_info.json":
            return mock_structure_info
        else:
            raise ValueError

    mock_loadfn = mocker.patch("pydefect.cli.vasp.main_vasp_functions.loadfn",
                               side_effect=side_effect)

    args = Namespace(dirs=[Path("Va_O1_2")],
                     p_state=mock_p_state,
                     y_range=[0.0, 1.0],
                     verbose=False)
    make_band_edge_orb_infos_and_eigval_plot(args)

    mock_procar.assert_called_with(Path("Va_O1_2") / defaults.procar)
    mock_vasprun.assert_called_with(Path("Va_O1_2") / defaults.vasprun)

    mock_make_orbital_infos.assert_called_with(
        mock_procar.return_value, mock_vasprun.return_value, 10, 20,
        mock_structure_info)

    mock_loadfn.assert_any_call(Path("Va_O1_2") / "defect_entry.json")

    mock_eigval_plotter.assert_called_with(
        title="Va_O1",
        band_edge_orb_infos=mock_make_orbital_infos.return_value,
        supercell_vbm=10,
        supercell_cbm=20,
        y_range=[0.0, 1.0])



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from typing import List, Union

import numpy as np
from numpy.linalg import inv
from pydefect.input_maker.supercell_info import SupercellInfo, Interstitial
from pydefect.util.error_classes import NotPrimitiveError
from pymatgen.core import Structure, Element, IStructure
from vise.util.structure_symmetrizer import StructureSymmetrizer
from vise.util.typing import Coords


def append_interstitial(supercell_info: SupercellInfo,
                        unitcell_structure: Union[Structure, IStructure],
                        frac_coords: List[Union[List[float], Coords]],
                        infos: List[str]
                        ) -> SupercellInfo:
    """
    inv_trans_mat must be multiplied with coords from the right as the
    trans_mat is multiplied to the unitcell lattice vector from the left.
    see __mul__ of IStructure in pymatgen.
    x_u, x_s means the frac coordinates in unitcell and supercell,
    while a, b, c are the unitcell lattice vector.
    (a_u, b_u, c_u) . (a, b, c) = (a_s, b_s, c_s) . trans_mat . (a, b, c)
    (a_u, b_u, c_u) = (a_s, b_s, c_s) . trans_mat
    so, (a_s, b_s, c_s) = (a_u, b_u, c_u) . inv_trans_ma
    """
    if supercell_info.unitcell_structure and \
            supercell_info.unitcell_structure != unitcell_structure:
        raise NotPrimitiveError

    if isinstance(frac_coords[0], float):
        frac_coords = [frac_coords]

    for fcoord, info in zip(frac_coords, infos):
        us = Structure.from_dict(unitcell_structure.as_dict())
        us.append(species=Element.H, coords=fcoord)
        symmetrizer = StructureSymmetrizer(us)
        site_symm = symmetrizer.spglib_sym_data["site_symmetry_symbols"][-1]

        inv_matrix = inv(np.array(supercell_info.transformation_matrix))
        new_coords = np.dot(fcoord, inv_matrix).tolist()

        supercell_info.interstitials.append(
            Interstitial(frac_coords=new_coords,
                         site_symmetry=site_symm,
                         info=info))
    return supercell_info



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

import numpy as np
import pytest

from pydefect.analyzer.unitcell import Unitcell
from vise.tests.helpers.assertion import assert_yaml_roundtrip


@pytest.fixture
def unitcell():
    ele = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]
    ion = [[10.0, 20.0, 30.0], [40.0, 50.0, 60.0], [70.0, 80.0, 90.0]]
    return Unitcell(system="MgAl2O4", vbm=-1.5053, cbm=5.1,
                    ele_dielectric_const=ele, ion_dielectric_const=ion)


def test_unitcell(unitcell):
    expected = np.array([[11, 22, 33], [44, 55, 66], [77, 88, 99]])
    np.testing.assert_array_almost_equal(unitcell.dielectric_constant, expected)


def test_yaml_roundtrip(unitcell, tmpdir):
    expected_text = """system: MgAl2O4
vbm: -1.5053
cbm: 5.1
ele_dielectric_const:
- - 1.0
  - 2.0
  - 3.0
- - 4.0
  - 5.0
  - 6.0
- - 7.0
  - 8.0
  - 9.0
ion_dielectric_const:
- - 10.0
  - 20.0
  - 30.0
- - 40.0
  - 50.0
  - 60.0
- - 70.0
  - 80.0
  - 90.0
"""
    assert_yaml_roundtrip(unitcell, tmpdir, expected_text)



# -*- coding: utf-8 -*-

from typing import List

from pydefect.defaults import defaults
from pymatgen.core import Element
from pymatgen.ext.matproj import MPRester

elements = [e.name for e in Element]


class MpQuery:
    def __init__(self,
                 element_list: List[str],
                 e_above_hull: float = defaults.e_above_hull,
                 properties: List[str] = None):
        default_properties = ["task_id", "full_formula", "final_energy",
                              "structure", "spacegroup", "band_gap",
                              "total_magnetization", "magnetic_type"]
        properties = properties or default_properties
        excluded = list(set(elements) - set(element_list))
        criteria = ({"elements": {"$in": element_list, "$nin": excluded},
                     "e_above_hull": {"$lte": e_above_hull}})
        # API key is parsed via .pmgrc.yaml
        with MPRester() as m:
            # Due to mp_decode=True by default, class objects are restored.
            self.materials = m.query(criteria=criteria,
                                     properties=properties)


class MpEntries:
    def __init__(self,
                 element_list: List[str],
                 e_above_hull: float = defaults.e_above_hull,
                 additional_properties: List[str] = None):
        excluded = list(set(elements) - set(element_list))
        criteria = ({"elements": {"$in": element_list, "$nin": excluded},
                     "e_above_hull": {"$lte": e_above_hull}})
        with MPRester() as m:
            self.materials = m.get_entries(
                criteria, property_data=additional_properties)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from dataclasses import dataclass
from typing import List

import numpy as np
from matplotlib import pyplot as plt
from monty.json import MSONable
from pydefect.defaults import defaults
from pydefect.util.coords import pretty_coords
from pymatgen.electronic_structure.core import Spin
from tabulate import tabulate
from vise.util.mix_in import ToJsonFileMixIn
from vise.util.typing import Coords


@dataclass
class AveChargeDensityDist(MSONable, ToJsonFileMixIn):
    """Average charge density distribution."""
    charge_center: Coords
    radial_dist: List[float]


@dataclass
class DefectChargeInfo(MSONable, ToJsonFileMixIn):
    distance_bins: List[float]  # the last value is the radius
    band_idxs: List[int]
    charge_dists: List[List[AveChargeDensityDist]]  # [band-idx, spin]
    ave_charge_density: float

    def ave_chg_dens_distribution(self, band_idx: int, spin: Spin):
        spin_idx = 0 if spin == Spin.up else 1
        band_pos = np.argwhere(np.array(self.band_idxs) == band_idx)[0][0]
        return self.charge_dists[band_pos][spin_idx].radial_dist

    def sum_chg_dens_distribution(self, band_idx: int, spin: Spin):
        ave = self.ave_chg_dens_distribution(band_idx, spin)
        volumes = (np.array(self.distance_bins[1:]) ** 3
                   - np.array(self.distance_bins[:-1]) ** 3) * 4 / 3 * np.pi
        return list(ave * volumes)

    @property
    def uniform_half_charge_radius(self):
        return (3 * 0.5 / (4 * np.pi * self.ave_charge_density)) ** (1.0 / 3.0)

    @property
    def is_spin_polarized(self):
        return len(self.charge_dists[0]) == 2

    def half_charge_radius(self, band_idx: int, spin: Spin):
        charge_sum = 0.0
        dist = self.sum_chg_dens_distribution(band_idx, spin)
        for c, start, end in zip(
                dist, self.distance_bins[:-1], self.distance_bins[1:]):
            charge_sum += c
            if charge_sum > 0.5:
                return start + (end - start) * (c + 0.5 - charge_sum) / c

        raise ValueError("Radius containing 0.5 e- could not be found.")

    def __str__(self):
        uniform_radius = f"{self.uniform_half_charge_radius:6.3f}"
        all_lines = [" -- defect charge info",
                     f"Uniform charge radius is {uniform_radius}"]
        lines = [["Band index", "Spin", "Radius", "Center"]]
        for band_idx, c_dist in zip(self.band_idxs, self.charge_dists):
            self._add_band_info(band_idx, c_dist, lines, Spin.up)
            if self.is_spin_polarized:
                self._add_band_info(band_idx, c_dist, lines, Spin.down)

        all_lines.append(tabulate(lines, tablefmt="plain"))
        return "\n".join(all_lines)

    def _add_band_info(self, band_idx, c_dist, lines, spin):
        try:
            radius = f"{self.half_charge_radius(band_idx, spin):6.3f}"
        except ValueError:
            radius = "None"
        spin_idx = 0 if spin == Spin.up else 1
        spin_str = "up" if spin == Spin.up else "down"
        center = pretty_coords(c_dist[spin_idx].charge_center)
        lines.append([band_idx + 1, spin_str, radius, center])

    @property
    def bins_middle_points(self):
        result = []
        for i, j in zip(self.distance_bins[:-1], self.distance_bins[1:]):
            result.append((i + j) / 2)
        return result

    def show_dist(self):
        plt.xlim([0, self.distance_bins[-1]])
        for charge_dist, _idx in zip(self.charge_dists, self.band_idxs):
            for dist, spin in zip(charge_dist, ["up", "down"]):
                plt.plot(self.bins_middle_points[:-1],
                         dist.radial_dist[:-1], label=f"{_idx} {spin}")
        ax = plt.gca()
        ax.legend(loc='upper right')
        return plt

    def localized_orbitals(self,
                           radius: float = defaults.localized_orbital_radius,
                           fraction_wrt_uniform: float
                           = defaults.localized_orbital_fraction_wrt_uniform):
        result = [[] for i in range(len(self.charge_dists[0]))]
        for charge_dist, band_idx in zip(self.charge_dists, self.band_idxs):
            for spin_idx, (dist, spin) in enumerate(zip(charge_dist,
                                                        [Spin.up, Spin.down])):
                try:
                    half_radius = self.half_charge_radius(band_idx, spin)
                    half_f = half_radius / self.uniform_half_charge_radius
                    if half_radius < radius and half_f < fraction_wrt_uniform:
                        result[spin_idx].append(band_idx)
                except ValueError:
                    pass

        return result



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
from monty.serialization import loadfn
from pydefect.analyzer.unitcell import Unitcell
from pydefect.cli.vasp.make_gkfo_correction import make_gkfo_correction
from pydefect.corrections.efnv_correction import \
    ExtendedFnvCorrection
from pydefect.corrections.gkfo_correction import GkfoCorrection


def test_make_gkfo_correction(vasp_files):
    d = vasp_files / "MgO_VaO_+_to_2+_transition"
    efnv_corr: ExtendedFnvCorrection = loadfn(d / "initial_+" / "correction.json")
    unitcell: Unitcell = loadfn(d / "unitcell.json")
    initial_calc_results = loadfn(d / "initial_+" / "calc_results.json")
    final_calc_results = loadfn(d / "final_2+" / "calc_results.json")

    gkfo = make_gkfo_correction(
        efnv_correction=efnv_corr,
        additional_charge=1,
        final_calc_results=final_calc_results,
        initial_calc_results=initial_calc_results,
        diele_tensor=unitcell.dielectric_constant,
        ion_clamped_diele_tensor=unitcell.ele_dielectric_const)

    np.testing.assert_almost_equal(gkfo.pc_1st_term, 0.5578636006366229)
    np.testing.assert_almost_equal(gkfo.pc_2nd_term, 0.883976881189451)
    np.testing.assert_almost_equal(gkfo.alignment_1st_term, -0.0743015221926139)
    np.testing.assert_almost_equal(gkfo.alignment_2nd_term, -0.1299646963025651)
    np.testing.assert_almost_equal(gkfo.alignment_3rd_term, -0.0234452481649620)
    np.testing.assert_almost_equal(gkfo.correction_energy, 1.2141290151659327)

    gkfo.to_json_file("gkfo_correction.json")







# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from copy import copy

import pytest
from pydefect.analyzer.calc_summary import SingleCalcSummary, CalcSummary
from pydefect.analyzer.defect_structure_info import DefectType, SymmRelation
from vise.tests.helpers.assertion import assert_msonable


@pytest.fixture
def single_summary():
    return SingleCalcSummary(charge=0,
                             atom_io={"O": -1},
                             electronic_conv=True,
                             ionic_conv=True,
                             is_energy_strange=False,
                             same_config_from_init=True,
                             defect_type=str(DefectType.vacancy),
                             symm_relation=str(SymmRelation.same),
                             donor_phs=False,
                             acceptor_phs=False,
                             unoccupied_deep_state=True,
                             occupied_deep_state=True)


def test_single_calc_summary_msonable(single_summary):
    assert_msonable(single_summary)


def test_single_calc_summary_properties(single_summary):
    other = copy(single_summary)
    assert single_summary.same_atom_charge_io(other) is True
    other.charge = 1
    assert single_summary.same_atom_charge_io(other) is False

    assert single_summary.is_converged is True
    assert single_summary.is_proper_result is True
    assert single_summary.is_unusual is False


@pytest.fixture
def calc_summary(single_summary):
    second = SingleCalcSummary(charge=1,
                               atom_io={"Mg": -1},
                               electronic_conv=True,
                               ionic_conv=False,
                               is_energy_strange=False,
                               same_config_from_init=False,
                               defect_type=str(DefectType.vacancy_split),
                               symm_relation=str(SymmRelation.supergroup),
                               donor_phs=True,
                               acceptor_phs=False,
                               unoccupied_deep_state=True,
                               occupied_deep_state=True)

    return CalcSummary({"Va_O1_0": single_summary, "Va_Mg1_1": second})


def test_calc_summary_msonable(calc_summary):
    assert_msonable(calc_summary)


def test_calc_summary_str(calc_summary):
    print(calc_summary)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import dash_table
from crystal_toolkit.core.mpcomponent import MPComponent
from crystal_toolkit.helpers.layouts import Column, dcc, html, Columns
from dash.dependencies import Input, Output
from pydefect.analyzer.defect_energy import DefectEnergySummary
from pydefect.analyzer.defect_energy_plotter import DefectEnergyPlotlyPlotter
from pydefect.chem_pot_diag.chem_pot_diag import ChemPotDiag, TargetVertices
from pydefect.chem_pot_diag.cpd_plotter import ChemPotDiag2DPlotlyPlotter, \
    ChemPotDiag3DPlotlyPlotter


def _target_vertices_to_columns_and_data(chem_pot_diag: ChemPotDiag):
    chem_pots = chem_pot_diag.to_target_vertices.chem_pots
    columns = [{"name": "label", "id": "label"}]
    for elem in chem_pot_diag.vertex_elements:
        columns.append({"name": elem, "id": elem})

    data = []
    for k, v in chem_pots.items():
        datum = {"label": k}
        for elem in chem_pot_diag.vertex_elements:
            datum[elem] = v[elem]
        data.append(datum)
    return columns, data


class CpdEnergyComponent(MPComponent):
    def __init__(self,
                 chem_pot_diag: ChemPotDiag,
                 defect_energy_summary: DefectEnergySummary,
                 *args, **kwargs):
        self.chem_pot_diag = chem_pot_diag
        self.dim = chem_pot_diag.dim
        self.defect_energy_summary = defect_energy_summary
        super().__init__(*args, **kwargs)

        columns, data = _target_vertices_to_columns_and_data(chem_pot_diag)
        self.vertex_list = dash_table.DataTable(
            id=self.id("vertex_list"),
            columns=columns,
            data=data,
            style_cell={'fontSize': 20, 'font-family': 'sans-serif',
                        "width": 700})

    def energy_fig(self, label, allow_shallow):
        plotter = DefectEnergyPlotlyPlotter(
            title=f"Label {label[0]}",
            chem_pot_label=label,
            defect_energy_summary=self.defect_energy_summary,
            allow_shallow=allow_shallow,
            with_corrections=True)
        return plotter.create_figure()

    @property
    def _sub_layouts(self):
        allow_shallow = self.get_choice_input(
            kwarg_label="allow_shallow",
            state={"label": "True", "value": "True"},
            label="Allow defects with perturbed host states",
            # Only string values are allowed.
            options=[{"label": "True", "value": "True"},
                     {"label": "False", "value": "False"}])

        if self.dim in [2, 3]:
            if self.dim == 2:
                fig = ChemPotDiag2DPlotlyPlotter(self.chem_pot_diag).figure
            else:
                fig = ChemPotDiag3DPlotlyPlotter(self.chem_pot_diag).figure
            cpd = dcc.Graph(id="cpd", figure=fig)
            energy = dcc.Graph(id=self.id('energy'))
            return {"cpd": cpd, "energy": energy, "allow_shallow": allow_shallow}
        else:
            d = [{"label": v, "value": v}
                 for v in self.chem_pot_diag.target_vertices_dict.keys()]
            cpd_label = self.get_choice_input(
                kwarg_label="label",
                state={"label": "A", "value": "A"},
                label="Equilibrium label",
                options=d)
            energy = dcc.Graph(id=self.id('energy'))
            return {"cpd_label": cpd_label, "energy": energy,
                    "allow_shallow": allow_shallow}

    @property
    def layout(self):
        if self.dim in [2, 3]:
            return Columns([Column([html.Div(self._sub_layouts["energy"]),
                                    self._sub_layouts["allow_shallow"]]),
                            Column([self._sub_layouts["cpd"],
                                    self.vertex_list])])

        else:
            return Columns([Column([self.vertex_list,
                                    self._sub_layouts["allow_shallow"],
                                    self._sub_layouts["cpd_label"]]),
                            Column(html.Div(self._sub_layouts["energy"]))])

    def generate_callbacks(self, app, cache):
        if self.dim in [2, 3]:
            @app.callback(
                Output(self.id('energy'), 'figure'),
                [Input('cpd', 'clickData'),
                 Input(self.get_kwarg_id('allow_shallow'), 'value')])
            def display_cpd(click_data, allow_shallow):
                allow_shallow = True if allow_shallow[0] == "True" else False
                label = click_data["points"][0]["text"] if click_data else "A"
                return self.energy_fig(label, allow_shallow)
        else:
            @app.callback(
                Output(self.id('energy'), 'figure'),
                [Input(self.get_kwarg_id("label"), "value"),
                 Input(self.get_kwarg_id('allow_shallow'), 'value')])
            def display_cpd(label, allow_shallow):
                allow_shallow = True if allow_shallow[0] == "True" else False
                label = label[0] if label else "A"
                return self.energy_fig(label, allow_shallow)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pydefect.input_maker.defect import SimpleDefect
from pydefect.input_maker.defect_set_maker import DefectSetMaker


def test_defect_set_maker(supercell_info):
    defect_maker = DefectSetMaker(supercell_info)
    expected = {SimpleDefect(None, "H1", [-1, 0, 1]),
                SimpleDefect(None, "He1", [0]),
                SimpleDefect("H", "i1", [-1, 0, 1]),
                SimpleDefect("He", "i1", [0])}
    assert defect_maker.defect_set == expected


def test_defect_set_maker_oxi_state(supercell_info):
    defect_maker = DefectSetMaker(supercell_info, {"H": 2})
    assert defect_maker.defect_set == {SimpleDefect(None, "H1", [-2, -1, 0]),
                                                           SimpleDefect(None, "He1", [0]),
                                                           SimpleDefect("H", "i1", [0, 1, 2]),
                                                           SimpleDefect("He", "i1", [0])}


def test_defect_set_maker_dopants(supercell_info):
    defect_maker = DefectSetMaker(supercell_info, dopants=["Li"])
    assert set(defect_maker.defect_set) == {
        SimpleDefect(None, "H1", [-1, 0, 1]),
        SimpleDefect(None, "He1", [0]),
        SimpleDefect("Li", "H1", [0]),
        SimpleDefect("H", "i1", [-1, 0, 1]),
        SimpleDefect("He", "i1", [0]),
        SimpleDefect("Li", "i1", [-1, 0, 1]),
    }


def test_defect_set_maker_electroneg_diff(supercell_info):
    defect_maker = DefectSetMaker(supercell_info, dopants=["Li"],
                                  ele_neg_diff=2.2 - 0.98 - 1e-5)
    assert set(defect_maker.defect_set) == {
        SimpleDefect(None, "H1", [-1, 0, 1]),
        SimpleDefect(None, "He1", [0]),
        SimpleDefect("H", "i1", [-1, 0, 1]),
        SimpleDefect("He", "i1", [0]),
        SimpleDefect("Li", "i1", [-1, 0, 1]),
    }


def test_defect_set_maker_screening(supercell_info):
    defect_maker = DefectSetMaker(supercell_info, keywords=["Va_H1"])
    expected = {SimpleDefect(None, "H1", [-1, 0, 1])}
    assert defect_maker.defect_set == expected





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
from monty.json import MSONable
from pydefect.corrections.abstract_correction import Correction
from tabulate import tabulate


@dataclass
class ExtendedFnvCorrection(Correction):
    """
    species: Species except for the defect. e.g., ["Mg", "Mg", ..., "O", ..]
    atomic_coords: Fractional coordinates except for the defect.
    pc_pot (list of float):
        List of point-charge potential from the defect for all the
        atomic sites.
    defect_region_radius (float):
        Maximum radius of a sphere touching to the lattice plane, used
        for defining the outside region of the defect.

    Add units of length and potential
    """
    charge: int
    point_charge_correction: float
    defect_region_radius: float
    sites: List["PotentialSite"]
    defect_coords: Tuple[float, float, float]

    def __str__(self):
        d = [["charge", self.charge],
             ["pc term", self.point_charge_correction],
             ["alignment term", self.alignment_correction],
             ["correction energy", self.correction_energy]]
        return tabulate(d, tablefmt='psql')

    @property
    def average_potential_diff(self):
        return np.mean([s.diff_pot for s in self.sites
                        if s.distance > self.defect_region_radius])

    @property
    def alignment_correction(self) -> float:
        return - self.average_potential_diff * self.charge

    @property
    def correction_energy(self) -> float:
        return self.point_charge_correction + self.alignment_correction

    @property
    def correction_dict(self):
        return {"pc term": self.point_charge_correction,
                "alignment term": self.alignment_correction}


@dataclass
class PotentialSite(MSONable):
    specie: str
    distance: float
    potential: float
    pc_potential: Optional[float]

    @property
    def diff_pot(self):
        return self.potential - self.pc_potential



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.calc_summary import SingleCalcSummary, CalcSummary
from pydefect.analyzer.defect_structure_info import DefectStructureInfo
from pydefect.analyzer.make_calc_summary import make_calc_summary
from pydefect.defaults import defaults
from pydefect.input_maker.defect_entry import DefectEntry
from pymatgen.core import IStructure, Lattice


def test_make_calc_summary(mocker):
    defect_entry = mocker.Mock(spec=DefectEntry, autospec=True)
    defect_entry.name = "Va_O1"
    defect_entry.charge = 1
    defect_entry.full_name = "Va_O1_1"

    calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    calc_results.structure = \
        IStructure(Lattice.cubic(1.0), ["Mg"], [[0.0]*3])
    calc_results.energy = 10.0 + defaults.abs_strange_energy - 0.1

    structure_info = mocker.Mock(spec=DefectStructureInfo, autospec=True)

    p_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    p_calc_results.structure = \
        IStructure(Lattice.cubic(1.0), ["Mg", "O"], [[0.0]*3]*2)
    p_calc_results.energy = 10.0

    actual = make_calc_summary(
        calc_set=[(calc_results, defect_entry, structure_info)],
        p_calc_results=p_calc_results)
    single_summary = SingleCalcSummary(
        charge=1,
        atom_io={"O": -1},
        electronic_conv=calc_results.electronic_conv,
        ionic_conv=calc_results.ionic_conv,
        is_energy_strange=False,
        same_config_from_init=structure_info.same_config_from_init,
        defect_type=str(structure_info.defect_type),
        symm_relation=str(structure_info.symm_relation))
    expected = CalcSummary({"Va_O1_1": single_summary})

    assert actual == expected





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

import argparse
import sys
import warnings
from pathlib import Path

from monty.serialization import loadfn
from pydefect import __version__
from pydefect.analyzer.unitcell import Unitcell
from pydefect.chem_pot_diag.chem_pot_diag import StandardEnergies
from pydefect.cli.main_functions import plot_defect_energy, \
    make_standard_and_relative_energies, make_cpd_and_vertices, \
    plot_chem_pot_diag, make_supercell, append_interstitial_to_supercell_info, \
    pop_interstitial_from_supercell_info, make_defect_set, \
    make_efnv_correction_main_func, make_band_edge_states_main_func, \
    make_defect_energy_infos_main_func, make_defect_energy_summary_main_func, \
    calc_defect_structure_info, make_calc_summary_main_func
from pydefect.cli.main_tools import str_int_to_int
from pydefect.defaults import defaults
from pymatgen.core import IStructure, Structure
from pymatgen.io.vasp.inputs import UnknownPotcarWarning

warnings.simplefilter('ignore', UnknownPotcarWarning)


description = """pydefect is a package that helps researchers to 
do first-principles point defect calculations with the VASP code."""

epilog = f"Author: Yu Kumagai Version: {__version__}"


def add_sub_parser(_argparse, name: str):
    result = _argparse.ArgumentParser(description="", add_help=False)
    if name == "dir":
        result.add_argument(
            "-d", "--dir", type=Path, required=True,
            help="Directory path to be parsed.")
    elif name == "dirs":
        result.add_argument(
            "-d", "--dirs", nargs="+", type=Path, required=True,
            help="Directory paths to be parsed.")
    elif name == "unitcell":
        result.add_argument(
            "-u", "--unitcell", type=Unitcell.from_yaml, required=True,
            help="Path to the unitcell.yaml file.")
    elif name == "supercell_info":
        result.add_argument(
            "-s", "--supercell_info", type=loadfn,
            default="supercell_info.json",
            help="Path to the supercell_info.json file.")
    elif name == "perfect_calc_results":
        result.add_argument(
            "-pcr", "--perfect_calc_results", required=True, type=loadfn,
            help="Path to the calc_results.json for the perfect supercell.")
    elif name == "perfect_band_edge_state":
        result.add_argument(
            "-pbes", "--p_state", required=True, type=loadfn,
            help="Path to the perfect_band_edge_state.json.")
    elif name == "no_calc_results_check":
        result.add_argument(
            "-nccr", "--no_calc_results_check",
            action="store_false",  dest="check_calc_results",
            help="Select this option when not checking calc_results.json.")
    elif name == "verbose":
        result.add_argument(
            "-v", "--verbose",
            action="store_true",  dest="verbose",
            help="Select if one wants to show traceback.")
    else:
        raise ValueError
    return result


dirs_parsers = [add_sub_parser(argparse, name="dirs"),
                add_sub_parser(argparse, name="verbose")]


def parse_args_main(args):

    parser = argparse.ArgumentParser(description=description, epilog=epilog)
    subparsers = parser.add_subparsers()

    unitcell_parser = add_sub_parser(argparse, name="unitcell")
    si_parser = add_sub_parser(argparse, name="supercell_info")
    pcr_parser = add_sub_parser(argparse, name="perfect_calc_results")
    pbes_parser = add_sub_parser(argparse, name="perfect_band_edge_state")
    no_calc_results = add_sub_parser(argparse, name="no_calc_results_check")

    # -- make_standard_and_relative_energies -----------------------------------
    parser_make_standard_and_relative_energies = subparsers.add_parser(
        name="standard_and_relative_energies",
        description="",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['sre'])

    parser_make_standard_and_relative_energies.add_argument(
        "-y", "--composition_energies_yaml", type=str,
        default="composition_energies.yaml",
        help="composition_energies.yaml file name.")
    parser_make_standard_and_relative_energies.set_defaults(
        func=make_standard_and_relative_energies)

    # -- make_cpd_and_vertices -------------------------------------------------
    parser_cv = subparsers.add_parser(
        name="cpd_and_vertices",
        description="Make chemical potential diagram.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['cv'])

    parser_cv.add_argument(
        "-y", "--rel_energy_yaml", type=str, default="relative_energies.yaml",
        help=" yaml file name.")
    parser_cv.add_argument(
        # this must be string as keys of RelativeEnergies are strings.
        "-t", "--target", type=str, required=True,
        help="Target composition, e.g., MgO.")
    parser_cv.add_argument(
        "-e", "--elements", type=str, nargs="+",
        help="Element names considered in chemical potential diagram. Used for "
             "creating the diagram.")
    parser_cv.set_defaults(func=make_cpd_and_vertices)

    # -- plot_cpd ------------------------------------------------
    parser_pcpd = subparsers.add_parser(
        name="plot_cpd",
        description="Plot chemical potential diagram and make pdf file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pc'])
    parser_pcpd.add_argument(
        "-cpd", "--chem_pot_diag", default="chem_pot_diag.json", type=loadfn,
        help="Path to the chem_pot_diag.json file.")
    parser_pcpd.set_defaults(func=plot_chem_pot_diag)

    # -- supercell ------------------------------------------------
    parser_supercell = subparsers.add_parser(
        name="supercell",
        description="Make SPOSCAR file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['s'])

    parser_supercell.add_argument(
        "-p", "--unitcell", type=IStructure.from_file, required=True,
        help="Base structure file, which must be the standardized primitive "
             "cell.")
    parser_supercell.add_argument(
        "--matrix", nargs="+", type=int,
        help="Supercell matrix applied to the conventional cell. "
             "1, 3 or 9 components are accepted.")
    parser_supercell.add_argument(
        "--min_atoms", dest="min_num_atoms", default=50, type=int,
        help="Minimum number of atoms")
    parser_supercell.add_argument(
        "--max_atoms", dest="max_num_atoms", default=300, type=int,
        help="Maximum number of atoms")

    parser_supercell.set_defaults(func=make_supercell)

    # -- append_interstitial ------------------------------------------------
    parser_append_interstitial = subparsers.add_parser(
        name="append_interstitial",
        description="Append interstitial information to supercell_info.yaml",
        parents=[si_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['ai'])

    parser_append_interstitial.add_argument(
        "-p", "--base_structure", required=True, type=Structure.from_file,
        help="Structure file defining the fractional coordinates such as the "
             "standardized primitive cell.")
    parser_append_interstitial.add_argument(
        "-c", "--frac_coords", required=True, nargs=3, type=float,
        help="Fractional coordinates in the standardized primitive cell added "
             "as an interstitial ")
    parser_append_interstitial.add_argument(
        "-i", "--info", type=str, default="None",
        help="Information related to the appended interstitial site if exists.")

    parser_append_interstitial.set_defaults(
        func=append_interstitial_to_supercell_info)

    # -- pop_interstitial ------------------------------------------------
    parser_pop_interstitial = subparsers.add_parser(
        name="pop",
        description="Pop interstitial site from supercell_info.json. ",
        parents=[si_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pi'])

    parser_pop_interstitial.add_argument(
        "-i", "--index", type=int,
        help="Popped interstitial index. Note that the indices begin from 1.")
    parser_pop_interstitial.add_argument(
        "--pop_all", action="store_true",
        help="Pop all interstitials. If this is set, index option is ignored.")

    parser_pop_interstitial.set_defaults(
        func=pop_interstitial_from_supercell_info)

    # -- defect_set ------------------------------------------------
    parser_defect_set = subparsers.add_parser(
        name="defect_set",
        description="Make defect_in.yaml file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['ds'])

    parser_defect_set.add_argument(
        "-o", "--oxi_states", nargs="+", type=str_int_to_int,
        help="Oxidation states in integers, e.g., Mg 2 O -2.")
    parser_defect_set.add_argument(
        "-d", "--dopants", nargs="+", type=str,
        help="Names of dopant element, e.g., Al Ga.")
    parser_defect_set.add_argument(
        "-k", "--keywords", nargs="+", type=str,
        help="Keywords used to screen the target defects. Since, the re.search "
             "is used inside, Regular expression can be used. ")

    parser_defect_set.set_defaults(func=make_defect_set)

    # -- defect structure info ------------------------------------------------
    parser_defect_structure_info = subparsers.add_parser(
        name="defect_structure_info",
        description="Defect structure info.",
        parents=dirs_parsers + [si_parser, no_calc_results],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['dsi'])

    parser_defect_structure_info.add_argument(
        "-dt", "--dist_tolerance", type=float, default=defaults.dist_tol,
        help="Tolerance in Angstrom for distance to judge if the "
             "atoms in initial and final structures are same ones.")
    parser_defect_structure_info.add_argument(
        "--symprec", type=float, default=defaults.symmetry_length_tolerance,
        help="Tolerance for determining point groups in the final "
             "structures. Note that point groups in the initial structures are "
             "set via defect_entry.json files.")
    parser_defect_structure_info.set_defaults(func=calc_defect_structure_info)

    # -- efnv correction ------------------------------------------------
    parser_efnv = subparsers.add_parser(
        name="efnv",
        description="Generate extended FNV-correction related efnv_correction "
                    "files. ",
        parents=dirs_parsers + [pcr_parser, unitcell_parser, no_calc_results],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['efnv'])

    parser_efnv.set_defaults(func=make_efnv_correction_main_func)

    # -- band edge states ------------------------------------------------
    parser_band_edge_states = subparsers.add_parser(
        name="band_edge_states",
        description="Show edge state for each spin channel.",
        parents=dirs_parsers + [pbes_parser, no_calc_results],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['bes'])

    parser_band_edge_states.set_defaults(func=make_band_edge_states_main_func)

    # -- defect energy infos ---------------------------------------------------
    parser_defect_energy_infos = subparsers.add_parser(
        name="defect_energy_infos",
        description="",
        parents=dirs_parsers + [unitcell_parser, pcr_parser, no_calc_results],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['dei'])
    parser_defect_energy_infos.add_argument(
        "-s", "--std_energies", required=True, type=StandardEnergies.from_yaml,
        help="Path to the StandardEnergies.yaml file.")

    parser_defect_energy_infos.set_defaults(
        func=make_defect_energy_infos_main_func)

    # -- defect energy summary -------------------------------------------------
    parser_defect_energy_summary = subparsers.add_parser(
        name="defect_energy_summary",
        description="",
        parents=dirs_parsers + [unitcell_parser, pbes_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['des'])
    parser_defect_energy_summary.add_argument(
        "-t", "--target_vertices_yaml", required=True, type=str,
        help="Path to the target_vertices.yaml file.")

    parser_defect_energy_summary.set_defaults(
        func=make_defect_energy_summary_main_func)

    # -- calc summary -------------------------------------------------
    parser_calc_summary = subparsers.add_parser(
        name="calc_summary",
        description="",
        parents=dirs_parsers + [pcr_parser, no_calc_results],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['cs'])

    parser_calc_summary.set_defaults(
        func=make_calc_summary_main_func)

    # -- plot defect formation energy ------------------------------------------
    parser_plot_energy = subparsers.add_parser(
        name="plot_defect_formation_energy",
        description="Show and plot defect formation energies.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pe'])

    parser_plot_energy.add_argument(
        "-d", "--defect_energy_summary", required=True, type=loadfn,
        help="defect_energy_summary.json file path.")
    parser_plot_energy.add_argument(
        "-l", "--label", type=str, help="Label")
    parser_plot_energy.add_argument(
        "-y", "--y_range", nargs=2, type=float,
        help="Energy range in y-axis")
    parser_plot_energy.add_argument(
        "--allow_shallow", action="store_true",
        help="Set when the energies of shallow defects are allowed.")
    parser_plot_energy.add_argument(
        "--no_corrections", dest="with_corrections", action="store_false",
        help="Set when corrections are switched off.")
    parser_plot_energy.add_argument(
        "--no_label_line", dest="label_line", action="store_false",
        help="Set the labels not to locate on the lines.")
    parser_plot_energy.add_argument(
        "--no_add_charges", dest="add_charges", action="store_false",
        help="Set the charges not to appear.")
    parser_plot_energy.set_defaults(func=plot_defect_energy)
    # ------------------------------------------------------------------------
    return parser.parse_args(args)


def main():
    args = parse_args_main(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.


class PydefectError(Exception):
    pass


class SupercellError(PydefectError):
    pass


class NotPrimitiveError(PydefectError):
    pass


class CpdNotSupportedError(Exception):
    pass



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
from monty.json import MSONable
from pydefect.defaults import defaults
from pydefect.util.structure_tools import Distances
from pymatgen.core import IStructure, Structure
from vise.util.typing import Coords


class DefectStructureComparator:
    def __init__(self,
                 defect_structure: IStructure,
                 perfect_structure: IStructure,
                 dist_tol: float = defaults.dist_tol):
        """
        Atoms in the final structure are shifted such that the farthest atom
        from the defect is placed at the same place with that in the perfect
        supercell.
        """
        self._defect_structure = defect_structure
        self._perfect_structure = perfect_structure
        self.dist_tol = dist_tol
        self.p_to_d = self.make_p_to_d()
        self.d_to_p = self.make_d_to_p()

    @property
    def atom_mapping(self):
        return {d: p for d, p in enumerate(self.d_to_p)
                if d not in self.inserted_indices}

    def _atom_projection(self, structure_from, structure_to, specie=True):
        result = []
        for site in structure_from:
            distances = Distances(structure_to,
                                  site.frac_coords,
                                  self.dist_tol)
            specie = site.specie if specie else None
            result.append(distances.atom_idx_at_center(specie=specie))
        return result

    def make_p_to_d(self):
        return self._atom_projection(
            self._perfect_structure, self._defect_structure)

    def make_d_to_p(self):
        return self._atom_projection(
            self._defect_structure, self._perfect_structure)

    @property
    def removed_indices(self):
        result = []
        for p, d in enumerate(self.p_to_d):
            try:
                if self.d_to_p[d] != p:
                    result.append(p)
            except (IndexError, TypeError):
                result.append(p)
        return sorted(result)

    @property
    def inserted_indices(self):
        result = []
        for d, p in enumerate(self.d_to_p):
            try:
                if self.p_to_d[p] != d:
                    result.append(d)
            except (IndexError, TypeError):
                result.append(d)
        return sorted(result)

    @property
    def defect_center_coord(self):
        coords = []
        for v in self.removed_indices:
            coords.append(self._perfect_structure[v].frac_coords)
        for i in self.inserted_indices:
            coords.append(self._defect_structure[i].frac_coords)

        lattice = self._perfect_structure.lattice
        repr_coords = coords[0]
        translated_coords = [list(repr_coords)]
        for c in coords[1:]:
            _, trans = lattice.get_distance_and_image(repr_coords, c)
            translated_coords.append([c[i] + trans[i] for i in range(3)])
        return np.average(translated_coords, axis=0) % 1

    def neighboring_atom_indices(self, cutoff_factor=None):
        distances = []
        for v in self.removed_indices:
            distances.append(Distances(self._defect_structure,
                                       self._perfect_structure[v].frac_coords,
                                       self.dist_tol))
        for i in self.inserted_indices:
            distances.append(Distances(self._defect_structure,
                                       self._defect_structure[i].frac_coords,
                                       self.dist_tol))
        result = set()
        for d in distances:
            result.update(d.coordination(cutoff_factor=cutoff_factor)
                          .neighboring_atom_indices)
        return sorted(list(result))

    def make_site_diff(self):
        removed_sites = [self._perfect_structure[x]
                         for x in self.removed_indices]
        inserted_sites = [self._defect_structure[x]
                          for x in self.inserted_indices]

        try:
            removed_str = Structure.from_sites(removed_sites)
        except ValueError:
            removed_str = None

        try:
            inserted_str = Structure.from_sites(inserted_sites)
        except ValueError:
            inserted_str = None

        if inserted_str and removed_str:
            r_to_i = self._atom_projection(removed_str, inserted_str, specie=False)
            i_to_r = self._atom_projection(inserted_str, removed_str, specie=False)
        else:
            r_to_i = [None]*len(removed_sites)
            i_to_r = [None]*len(inserted_sites)

        mapping, removed_mapping, inserted_mapping = [], [], []
        for x, y in enumerate(r_to_i):
            if y and x == i_to_r[y]:
                mapping.append((self.removed_indices[x], self.inserted_indices[y]))
                removed_mapping.append(self.removed_indices[x])
                inserted_mapping.append(self.inserted_indices[y])

        removed, removed_by_sub = [], []
        for idx in self.removed_indices:
            site = self._perfect_structure[idx]
            frac_coords = tuple([float(fc) for fc in site.frac_coords])
            val = idx, site.species_string, frac_coords
            if idx in removed_mapping:
                removed_by_sub.append(val)
            else:
                removed.append(val)

        inserted, inserted_by_sub = [], []
        for idx in self.inserted_indices:
            site = self._defect_structure[idx]
            frac_coords = tuple([float(fc) for fc in site.frac_coords])
            val = idx, site.species_string, frac_coords
            if idx in inserted_mapping:
                inserted_by_sub.append(val)
            else:
                inserted.append(val)

        return SiteDiff(removed=removed,
                        inserted=inserted,
                        removed_by_sub=removed_by_sub,
                        inserted_by_sub=inserted_by_sub)


SiteInfo = Tuple[int, str, Coords]


@dataclass
class SiteDiff(MSONable):
    removed: List[SiteInfo]
    inserted: List[SiteInfo]
    removed_by_sub: List[SiteInfo]
    inserted_by_sub: List[SiteInfo]

    @classmethod
    def from_dict(cls, d):
        result = super().from_dict(d)
        removed, inserted, removed_by_sub, inserted_by_sub = [], [], [], []

        for i in result.removed:
            removed.append((i[0], i[1], tuple(i[2])))
        for i in result.inserted:
            inserted.append((i[0], i[1], tuple(i[2])))
        for i in result.removed_by_sub:
            removed_by_sub.append((i[0], i[1], tuple(i[2])))
        for i in result.inserted_by_sub:
            inserted_by_sub.append((i[0], i[1], tuple(i[2])))

        return cls(removed, inserted, removed_by_sub, inserted_by_sub)

    @property
    def is_complex_defect(self):
        return (len(self.removed) + len(self.inserted)
                + len(self.removed_by_sub)) != 1

    @property
    def is_vacancy(self):
        return (len(self.removed) == 1 and len(self.inserted) == 0 and
                len(self.removed_by_sub) == 0 and len(self.inserted_by_sub) == 0)

    @property
    def is_interstitial(self):
        return (len(self.removed) == 0 and len(self.inserted) == 1 and
                len(self.removed_by_sub) == 0 and len(self.inserted_by_sub) == 0)

    @property
    def is_substituted(self):
        return (len(self.removed) == 0 and len(self.inserted) == 0 and
                len(self.removed_by_sub) == 1 and len(self.inserted_by_sub) == 1)

    @property
    def is_no_diff(self):
        return not (self.removed or self.inserted
                    or self.removed_by_sub or self.inserted_by_sub)




# coding: utf-8
#  Copyright (c) 2020 Kumagai group.
import argparse
import sys
from pathlib import Path

import crystal_toolkit.components as ctc
from ace.htmls.components.supercell_component import \
    SupercellComponent
from crystal_toolkit.helpers.layouts import *
from dash import Dash
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.dash_components.cpd_energy_dash import \
    CpdEnergy2D3DComponent, CpdEnergyOtherComponent
from pydefect.analyzer.dash_components.scenes_from_volumetric_data import \
    SceneDicts
from pydefect.analyzer.dash_components.single_defect_component import \
    SingleDefectComponent
from pydefect.analyzer.eigenvalue_plotter import EigenvaluePlotlyPlotter
from pydefect.chem_pot_diag.chem_pot_diag import CpdPlotInfo, ChemPotDiag
from pydefect.corrections.site_potential_plotter import \
    SitePotentialPlotlyPlotter
from pydefect.defaults import defaults
from pydefect.input_maker.defect_entry import DefectEntry
from pymatgen.core import Structure
from vise.analyzer.dash_components.band_dos_dash import BandDosComponent
from vise.analyzer.dash_components.main import symmetry_layout, site_layout, \
    mpid_and_link
from vise.analyzer.dash_components.structure_component import StructureComponent
from vise.util.structure_symmetrizer import StructureSymmetrizer

app = Dash(suppress_callback_exceptions=True,
           assets_folder=SETTINGS.ASSETS_PATH)


def create_ctk(struct_component,
               crystal_symmetry,
               mpid_and_link,
               sites,
               band_dos_component,
               supercell_component_layout,
               cpd_energy_layout,
               single_defect_layouts):
    box_size = "30vmin"
    supercell = Column(
        [
            Box(
                supercell_component_layout,
                style={
                    "width": box_size,
                    "height": box_size,
                    "minWidth": "300px",
                    "minHeight": "300px",
                    "maxWidth": "600px",
                    "maxHeight": "600px",
                    "overflow": "hidden",
                    "padding": "0.25rem",
                    "marginBottom": "0.5rem",
                },
            ),
        ],
        narrow=True,
    )


    _layout = Container(
        [
            Section(
                [
                    Columns(
                        [
                            Column(
                                [struct_component.title_layout()]
                            )
                        ]
                    ),
                    Columns(
                        [
                            Column(
                                [
                                    Box(
                                        struct_component.layout(size="100%"),
                                        style={
                                            "width": box_size,
                                            "height": box_size,
                                            "minWidth": "300px",
                                            "minHeight": "300px",
                                            "maxWidth": "800px",
                                            "maxHeight": "800px",
                                            "overflow": "hidden",
                                            "padding": "0.25rem",
                                            "marginBottom": "0.5rem",
                                        },
                                    ),
                                    html.Div(
                                        [
                                            html.Div(
                                                struct_component.legend_layout(),
                                                style={"float": "left"},
                                            ),
                                        ],
                                        style={
                                            "width": box_size,
                                            "minWidth": "300px",
                                            "marginBottom": "40px",
                                        },
                                    ),
                                ],
                                narrow=True,
                            ),
                            Column(
                                [crystal_symmetry, mpid_and_link],
                                style={"width": box_size, "max-width": box_size},
                            ),
                        ],
                        desktop_only=False,
                        centered=False,
                    ),
                    Columns(Column([sites])),
                ]
            ),
            Section([band_dos_component.layout()]
                    ),
            Section([Columns(cpd_energy_layout + [supercell]),
                     Column(single_defect_layouts)])
        ]
    )

    return _layout


def make_layouts(structure: Structure, dos_plot_data, band_plot_data,
                 perfect_dirname, defect_dirnames, supercell_info, chem_pot_diag):
    cpd_plot_info = CpdPlotInfo(ChemPotDiag.from_yaml(chem_pot_diag))
    perfect: CalcResults = loadfn(perfect_dirname / "calc_results.json")

    defects, defect_entries, corrections, single_defect_layouts = [], [], [], []
    for d in defect_dirnames:
        #TODO: check convergence
        defects.append(loadfn(d / "calc_results.json"))
        defect_entry = loadfn(d / "defect_entry.json")
        defect_entries.append(defect_entry)
        corrections.append(loadfn(d / "correction.json"))
        defect_entry: DefectEntry = loadfn(d / "defect_entry.json")
        efnv_correction = loadfn(d / "correction.json")
        eigval = loadfn(d / "band_edge_eigenvalues.json")

        pot = SitePotentialPlotlyPlotter(title=defect_entry.full_name,
                                         efnv_correction=efnv_correction)
        eig = EigenvaluePlotlyPlotter(title=defect_entry.full_name,
                                      band_edge_eigenvalues=eigval,
                                      supercell_vbm=perfect.vbm,
                                      supercell_cbm=perfect.cbm)
        scene_dicts = loadfn(d / "parchgs" / "scene_dicts.json")
        if isinstance(scene_dicts, dict):
            scene_dicts = SceneDicts.from_dict(scene_dicts)

        single_defect_layouts.append(
            SingleDefectComponent(pot, eig, scene_dicts, defect_entry.full_name,
                                  id=f"{defect_entry.full_name}").layout())

    if cpd_plot_info.cpd.dim in [2, 3]:
        cpd_energy_comp = CpdEnergy2D3DComponent(cpd_plot_info, perfect, defects,
                                             defect_entries, corrections)
    else:
        cpd_energy_comp = CpdEnergyOtherComponent(cpd_plot_info, perfect, defects,
                                                  defect_entries, corrections)

    structure_component = StructureComponent(structure)
    comp = structure.composition.reduced_formula
    band_dos_component = BandDosComponent(dos_plot_data, band_plot_data, id=f"band_dos_{comp}",)
    supercell_component = SupercellComponent(supercell_info)
    symmetrizer = StructureSymmetrizer(structure)
    return create_ctk(structure_component,
                      symmetry_layout(structure),
                      mpid_and_link(symmetrizer),
                      site_layout(symmetrizer),
                      band_dos_component,
                      supercell_component.layout,
                      cpd_energy_comp.layout,
                      single_defect_layouts)


def parse_args(args):
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-b", "--band", type=Path, required=True)
    parser.add_argument("-d", "--dos", type=Path, required=True)
    parser.add_argument("-p", "--perfect", type=Path, required=True)
    parser.add_argument("-def", "--defects", type=Path, nargs="+", required=True)
    parser.add_argument("-s", "--supercell_info", type=loadfn, required=True)
    parser.add_argument("-c", "--chem_pot_diag", type=str, required=True)
    parser.add_argument("--port", type=int)
    return parser.parse_args(args)


if __name__ == "__main__":
    args = parse_args(sys.argv[1:])
    structure = Structure.from_file(args.dos / defaults.contcar)
    dos_plot_data = loadfn(args.dos / "dos_plot_data.json")
    band_plot_data = loadfn(args.band / "band_plot_info.json")
    layout = make_layouts(structure, dos_plot_data, band_plot_data,
                          args.perfect, args.defects, args.supercell_info,
                          args.chem_pot_diag)
    ctc.register_crystal_toolkit(app=app, layout=layout, cache=None)
    app.run_server(debug=True, port=args.port)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import deepcopy

from pydefect.analyzer.defect_structure_comparator import SiteDiff
from pydefect.analyzer.defect_structure_info import \
    symmetry_relation, SymmRelation, \
    DefectType, judge_defect_type, unique_point_group
from pymatgen.core import Structure, Lattice
from vise.tests.helpers.assertion import assert_msonable, assert_json_roundtrip
from vise.util.enum import ExtendedEnum
from vise.util.structure_symmetrizer import num_sym_op


def test_defect_type():
    assert_msonable(DefectType.vacancy)
    assert issubclass(DefectType, ExtendedEnum)


def test_judge_defect_type():
    """The structures remain as they could be used in the future"""
    perfect = Structure(Lattice.cubic(10), species=["H"]*4,
                        coords=[[0.0, 0.0, 0.0], [0.0, 0.5, 0.5],
                                [0.5, 0.0, 0.5], [0.5, 0.5, 0.0]])
    defect = Structure(Lattice.cubic(10), species=["H"]*3,
                       coords=[[0.0, 0.0, 0.0], [0.25, 0.25, 0.5],
                               [0.5, 0.5, 0.0]])
    site_diff = SiteDiff(removed=[(1, "H", (0.0, 0.5, 0.5)),
                                  (2, "H", (0.5, 0.0, 0.5))],
                         inserted=[(1, "H", (0.25, 0.25, 0.5))],
                         removed_by_sub=[], inserted_by_sub=[])
    assert judge_defect_type(site_diff) == DefectType.vacancy_split

    defect_2 = Structure(Lattice.cubic(10), species=["H"]*5,
                         coords=[[0.0, 0.0, 0.0], [0.25, 0.25, 0.5],
                                 [0.75, 0.25, 0.5], [0.25, 0.75, 0.5],
                                 [0.5, 0.5, 0.0]])
    site_diff_2 = SiteDiff(removed=[(1, "H", (0.0, 0.5, 0.5)),
                                    (2, "H", (0.5, 0.0, 0.5))],
                           inserted=[(1, "H", (0.25, 0.25, 0.5)),
                                     (2, "H", (0.75, 0.25, 0.5)),
                                     (3, "H", (0.25, 0.75, 0.5))],
                           removed_by_sub=[], inserted_by_sub=[])
    assert judge_defect_type(site_diff_2) == DefectType.interstitial_split

    defect_3 = Structure(Lattice.cubic(10), species=["H", "He", "H"],
                         coords=[[0.0, 0.0, 0.0], [0.25, 0.25, 0.5],
                                 [0.5, 0.5, 0.0]])
    site_diff_3 = SiteDiff(removed=[(1, "H", (0.0, 0.5, 0.5)),
                                    (2, "H", (0.5, 0.0, 0.5))],
                           inserted=[(1, "He", (0.25, 0.25, 0.5))],
                           removed_by_sub=[], inserted_by_sub=[])

    assert judge_defect_type(site_diff_3) == DefectType.unknown


def test_defect_structure_info_json_roundtrip(def_str_info, tmpdir):
    assert_json_roundtrip(def_str_info, tmpdir)


def test_symmetry_relation():
    for point_group in num_sym_op:
        pg = unique_point_group(point_group)
        assert symmetry_relation(pg, pg) == SymmRelation.same
    assert symmetry_relation("3m", "m") == SymmRelation.subgroup
    assert symmetry_relation("4", "2") == SymmRelation.subgroup
    assert symmetry_relation("-43m", "mm2") == SymmRelation.subgroup
    assert symmetry_relation("4/m", "2") == SymmRelation.subgroup


def test_make_def_str_info_symm_rel(def_str_info):
    assert def_str_info.symm_relation == SymmRelation.subgroup


def test_make_def_str_info_same_config_from_init(def_str_info):
    assert def_str_info.same_config_from_init is False


def test_def_str_info_defect_type(def_str_info):
    assert def_str_info.defect_type == DefectType.unknown


def test_repr(def_str_info):
    x = deepcopy(def_str_info)
    x.site_diff_from_initial.inserted = []
    actual = x.__str__()
    expected = """ -- defect structure info
Defect type: unknown
Site symmetry: 3m -> m (subgroup)
Has same configuration from initial structure: False
Drift distance: 0.001
Defect center: ( 0.380,  0.375,  0.375)
Removed atoms:
0  H  ( 0.250,  0.250,  0.250)

Added atoms:
0  H  ( 0.270,  0.250,  0.250)

Removed atoms from initial structure:
0  H  ( 0.250,  0.250,  0.250)

Neighbor max distance 2.563
Displacements
Elem  Dist  Displace  Angle  Index  Initial site                  Final site                Neighbor
Be    2.14  0.01      130    2      ( 0.500,  0.500,  0.500)  ->  ( 0.500,  0.500,  0.501)"""
    assert actual == expected





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict

from monty.json import MSONable
from monty.serialization import loadfn


class Correction(ABC, MSONable):
    def to_json_file(self, filename: str = "correction.json") -> None:
        Path(filename).write_text(self.to_json())

    @classmethod
    def from_json_file(cls, filename: str = "correction.json"):
        return loadfn(filename)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pathlib import Path
from typing import Union

from monty.serialization import loadfn
from pydefect.analyzer.calc_results import CalcResults, NoElectronicConvError, \
    NoIonicConvError
from pydefect.analyzer.defect_energy import DefectEnergyInfo
from pydefect.analyzer.defect_energy_plotter import DefectEnergyMplPlotter
from pydefect.analyzer.make_band_edge_states import make_band_edge_states
from pydefect.analyzer.make_calc_summary import make_calc_summary
from pydefect.analyzer.make_defect_energy_info import make_defect_energy_info
from pydefect.analyzer.make_defect_energy_summary import \
    make_defect_energy_summary
from pydefect.analyzer.make_defect_structure_info import \
    MakeDefectStructureInfo
from pydefect.chem_pot_diag.chem_pot_diag import CompositionEnergies, \
    RelativeEnergies, ChemPotDiagMaker, TargetVertices, change_element_sequence
from pydefect.chem_pot_diag.cpd_plotter import ChemPotDiag2DMplPlotter, \
    ChemPotDiag3DMplPlotter
from pydefect.cli.main_tools import sanitize_matrix, parse_dirs
from pydefect.cli.vasp.make_efnv_correction import make_efnv_correction
from pydefect.corrections.site_potential_plotter import SitePotentialMplPlotter
from pydefect.input_maker.append_interstitial import append_interstitial
from pydefect.input_maker.defect_set_maker import DefectSetMaker
from pydefect.input_maker.supercell_maker import SupercellMaker
from pymatgen.core import Composition
from vise.util.logger import get_logger

logger = get_logger(__name__)


def get_calc_results(d: Path, check: bool) -> Union[CalcResults, bool]:
    if check:
        try:
            calc_results = loadfn(d / "calc_results.json")
        except FileNotFoundError:
            logger.warning(f"calc_results.json doesn't exist in {d}.")
            raise

        if calc_results.electronic_conv is False:
            logger.warning(f"SCF in {d} is not reached.")
            raise NoElectronicConvError
        elif calc_results.ionic_conv is False:
            logger.warning(f"Ionic convergence in {d} is not reached.")
            raise NoIonicConvError
    else:
        calc_results = loadfn(d / "calc_results.json")

    return calc_results


def make_standard_and_relative_energies(args):
    comp_energies = CompositionEnergies.from_yaml(
        args.composition_energies_yaml)
    std_energies, rel_energies = comp_energies.std_rel_energies
    std_energies.to_yaml_file()
    rel_energies.to_yaml_file()


def make_cpd_and_vertices(args):
    rel_energies = RelativeEnergies.from_yaml(args.rel_energy_yaml)
    if args.target:
        elements = args.elements or \
                   Composition(args.target).chemical_system.split("-")
    else:
        elements = args.elements or list(rel_energies.all_element_set)

    cpd_maker = ChemPotDiagMaker(rel_energies, elements, args.target)
    cpd = cpd_maker.chem_pot_diag
    cpd.to_json_file()
    cpd.to_target_vertices.to_yaml_file()


def plot_chem_pot_diag(args):
    cpd = args.chem_pot_diag
    cpd = change_element_sequence(cpd)
    if cpd.dim == 2:
        plotter = ChemPotDiag2DMplPlotter(cpd)
    elif cpd.dim == 3:
        plotter = ChemPotDiag3DMplPlotter(cpd)
    else:
        raise ValueError(f"Only 2 or 3 dimensions are supported. "
                         f"Now {cpd.dim} dimensions.")

    plt = plotter.draw_diagram()
    plt.savefig(fname="cpd.pdf")
    plt.show()


def make_supercell(args):
    if args.matrix:
        matrix = sanitize_matrix(args.matrix)
        maker = SupercellMaker(args.unitcell, matrix)
    else:
        kwargs = {}
        if args.min_num_atoms:
            kwargs["min_num_atoms"] = args.min_num_atoms
        if args.max_num_atoms:
            kwargs["max_num_atoms"] = args.max_num_atoms
        maker = SupercellMaker(args.unitcell, **kwargs)

    maker.supercell.structure.to(filename="SPOSCAR")
    maker.supercell_info.to_json_file()


def append_interstitial_to_supercell_info(args):
    supercell_info = append_interstitial(args.supercell_info,
                                         args.base_structure,
                                         [args.frac_coords],
                                         args.info)
    supercell_info.to_json_file()


def pop_interstitial_from_supercell_info(args):
    supercell_info = args.supercell_info
    logger.info("Be careful that the interstitials indices are changed.")
    if args.pop_all is False:
        assert args.index > 0
        supercell_info.interstitials.pop(args.index - 1)
    else:
        try:
            while True:
                supercell_info.interstitials.pop()
        except IndexError:
            pass

    supercell_info.to_json_file()


def make_defect_set(args):
    supercell_info = loadfn("supercell_info.json")
    if args.oxi_states:
        oxi_states = dict(zip(args.oxi_states[::2], args.oxi_states[1::2]))
    else:
        oxi_states = None
    maker = DefectSetMaker(supercell_info,
                           oxi_states,
                           args.dopants,
                           keywords=args.keywords)
    maker.defect_set.to_yaml()


def calc_defect_structure_info(args):
    supercell_info = args.supercell_info
    file_name = "defect_structure_info.json"

    def _inner(_dir: Path):
        calc_results = get_calc_results(_dir, args.check_calc_results)
        defect_entry = loadfn(_dir / "defect_entry.json")
        defect_str_info = MakeDefectStructureInfo(
            supercell_info.structure,
            defect_entry.structure,
            calc_results.structure,
            dist_tol=args.dist_tolerance,
            symprec=args.symprec).defect_structure_info
        defect_str_info.to_json_file(str(_dir / file_name))

    parse_dirs(args.dirs, _inner, args.verbose, file_name)


def make_efnv_correction_main_func(args):
    file_name = "correction.json"
    def _inner(_dir: Path):
        calc_results = get_calc_results(_dir, args.check_calc_results)
        defect_entry = loadfn(_dir / "defect_entry.json")
        efnv = make_efnv_correction(defect_entry.charge,
                                    calc_results,
                                    args.perfect_calc_results,
                                    args.unitcell.dielectric_constant)
        efnv.to_json_file(_dir / file_name)

        title = defect_entry.full_name
        plotter = SitePotentialMplPlotter.from_efnv_corr(
            title=title, efnv_correction=efnv)
        plotter.construct_plot()
        plotter.plt.savefig(fname=_dir / "correction.pdf")
        plotter.plt.clf()

    parse_dirs(args.dirs, _inner, args.verbose, file_name)


def make_band_edge_states_main_func(args):
    file_name = "band_edge_states.json"
    def _inner(_dir: Path):
        orb_infos = loadfn(_dir / "band_edge_orbital_infos.json")
        try:
            defect_charge_info = loadfn(_dir / "defect_charge_info.json")
        except FileNotFoundError:
            defect_charge_info = None
        band_edge_states = make_band_edge_states(orb_infos, args.p_state,
                                                 defect_charge_info)
        band_edge_states.to_json_file(str(_dir / file_name))

    parse_dirs(args.dirs, _inner, args.verbose, file_name)


def make_defect_energy_infos_main_func(args):
    file_name = "defect_energy_info.yaml"
    def _inner(_dir: Path):
        calc_results = get_calc_results(_dir, args.check_calc_results)
        defect_entry = loadfn(_dir / "defect_entry.json")
        correction = loadfn(_dir / "correction.json")
        try:
            band_edge_states = loadfn(_dir / "band_edge_states.json")
        except FileNotFoundError:
            band_edge_states = None
        defect_energy_info = make_defect_energy_info(
            defect_entry=defect_entry,
            calc_results=calc_results,
            correction=correction,
            perfect_calc_results=args.perfect_calc_results,
            standard_energies=args.std_energies,
            unitcell=args.unitcell,
            band_edge_states=band_edge_states)
        defect_energy_info.to_yaml_file(str(_dir / file_name))

    parse_dirs(args.dirs, _inner, args.verbose, file_name)


def make_defect_energy_summary_main_func(args):
    def _inner(_dir: Path):
        return DefectEnergyInfo.from_yaml(str(_dir / "defect_energy_info.yaml"))

    energy_infos = parse_dirs(args.dirs, _inner, args.verbose)
    target_vertices = TargetVertices.from_yaml(args.target_vertices_yaml)
    defect_energy_summary = make_defect_energy_summary(
        energy_infos, target_vertices, args.unitcell, args.p_state)
    defect_energy_summary.to_json_file()


def make_calc_summary_main_func(args):
    def _inner(_dir: Path):
        calc_results = get_calc_results(_dir, args.check_calc_results)
        defect_entry = loadfn(_dir / "defect_entry.json")
        str_info = loadfn(_dir / "defect_structure_info.json")
        return calc_results, defect_entry, str_info

    _infos = parse_dirs(args.dirs, _inner, args.verbose)
    calc_summary = make_calc_summary(_infos, args.perfect_calc_results)
    calc_summary.to_json_file()


def plot_defect_energy(args):
    plotter = DefectEnergyMplPlotter(
        defect_energy_summary=args.defect_energy_summary,
        chem_pot_label=args.label,
        y_range=args.y_range,
        allow_shallow=args.allow_shallow,
        with_corrections=args.with_corrections,
        label_line=args.label_line,
        add_charges=args.add_charges)
    plotter.construct_plot()
    plotter.plt.savefig(f"energy_{args.label}.pdf")



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from typing import Dict

import numpy as np
from pydefect.analyzer.band_edge_states import OrbitalInfo, \
    BandEdgeOrbitalInfos
from pydefect.analyzer.defect_structure_info import DefectStructureInfo
from pydefect.defaults import defaults
from pymatgen.core import Structure
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Procar, Vasprun


def make_band_edge_orbital_infos(procar: Procar,
                                 vasprun: Vasprun,
                                 vbm: float, cbm: float,
                                 str_info: DefectStructureInfo):
    eigval_range = defaults.eigval_range
    kpt_coords = [tuple(coord) for coord in vasprun.actual_kpoints]
    max_energy_by_spin, min_energy_by_spin = [], []
    neighbors = str_info.neighbor_atom_indices

    for e in vasprun.eigenvalues.values():
        max_energy_by_spin.append(np.amax(e[:, :, 0], axis=0))
        min_energy_by_spin.append(np.amin(e[:, :, 0], axis=0))

    max_energy_by_band = np.amax(np.vstack(max_energy_by_spin), axis=0)
    min_energy_by_band = np.amin(np.vstack(min_energy_by_spin), axis=0)

    lower_idx = np.argwhere(max_energy_by_band > vbm - eigval_range)[0][0]
    upper_idx = np.argwhere(min_energy_by_band < cbm + eigval_range)[-1][-1]

    orbs, s = procar.data, vasprun.final_structure
    orb_infos = []
    for spin, eigvals in vasprun.eigenvalues.items():
        orb_infos.append([])
        for k_idx in range(len(kpt_coords)):
            orb_infos[-1].append([])
            for b_idx in range(lower_idx, upper_idx + 1):
                e, occ = eigvals[k_idx, b_idx, :]
                orbitals = calc_orbital_character(orbs, s, spin, k_idx, b_idx)
                p_ratio = calc_participation_ratio(orbs, spin, k_idx, b_idx, neighbors)
                orb_infos[-1][-1].append(OrbitalInfo(e, orbitals, occ, p_ratio))

    return BandEdgeOrbitalInfos(
        orbital_infos=orb_infos,
        kpt_coords=kpt_coords,
        kpt_weights=vasprun.actual_kpoints_weights,
        # need to convert numpy.int64 to int for mongoDB.
        lowest_band_index=int(lower_idx),
        fermi_level=vasprun.efermi)


def calc_participation_ratio(orbitals: Dict[Spin, np.ndarray],
                             spin: Spin,
                             kpt_index: int,
                             band_index: int,
                             atom_indices: list) -> float:
    """ Returns sum of participation ratios at atom_indices sites

    The PROCAR data of the form below. It should VASP uses 1-based indexing,
    but all indices are converted to 0-based here.::
        { spin: np.ndarray accessed with (k-point index, band index,
                                          ion index, orbital index) }

    Note that the k-point weight is not considered, so all the k-points are
    treated equally.

    Return (float):
        float of the participation ratio.
    """
    sum_per_atom = np.sum(orbitals[spin][kpt_index, band_index, :, :], axis=1)
    return np.sum(sum_per_atom[atom_indices]) / np.sum(sum_per_atom)


def calc_orbital_character(orbitals,
                           structure: Structure,
                           spin: Spin,
                           kpt_index: int,
                           band_index: int):
    """ Consider the two pattern of orbitals.

    LORBIT 10 -> consider only "s", "p", "d" orbitals
    LORBIT >=11 -> consider "s", "px", "py", "pz",.. orbitals

    """

    def projection_sum(atom_indices: tuple, first: int, last: int):
        end = last + 1
        procar_sum = np.sum(orbitals[spin]
                            [kpt_index, band_index, atom_indices, first:end])
        return float(procar_sum)

    orbital_components = {}
    azimuthal = len(orbitals[Spin.up][0, 0, 0]) > 5

    for element in structure.symbol_set:
        # get list of index
        indices = structure.indices_from_symbol(element)
        if azimuthal:
            orbital_components[element] = \
                [round(projection_sum(indices, 0, 0), 3),
                 round(projection_sum(indices, 1, 3), 3),
                 round(projection_sum(indices, 4, 8), 3)]
            try:
                orbital_components[element].append(round(projection_sum(indices, 9, 16), 3))
            except KeyError:
                pass
        else:
            orbital_components[element] = \
                [round(projection_sum(indices, 0, 0), 3),
                 round(projection_sum(indices, 1, 1), 3),
                 round(projection_sum(indices, 2, 2), 3)]
    return orbital_components


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from typing import List, Tuple, Optional

import numpy as np
from pydefect.defaults import defaults
from pydefect.input_maker.defect import SimpleDefect
from pydefect.input_maker.defect_entry import DefectEntry, PerturbedSite
from pydefect.input_maker.defect_set import DefectSet
from pydefect.input_maker.supercell_info import SupercellInfo
from pymatgen.core import Structure, IStructure
from pymatgen.core.structure import PeriodicNeighbor
from vise.util.structure_symmetrizer import StructureSymmetrizer
from vise.util.typing import Coords


class DefectEntriesMaker:
    def __init__(self, supercell_info: SupercellInfo, defect_set: DefectSet):
        self.supercell_info = supercell_info
        self.defect_entries = set()

        for defect in defect_set:
            (structure, coords, site_sym,
             perturbed_structure, perturbed_sites, perturbed_site_sym) = \
                self._create_defect_structures(defect)

            for charge in defect.charges:
                self.defect_entries.add(
                    DefectEntry(name=defect.name,
                                charge=charge,
                                structure=structure,
                                site_symmetry=site_sym,
                                defect_center=coords,
                                perturbed_structure=perturbed_structure,
                                perturbed_sites=perturbed_sites,
                                perturbed_site_symmetry=perturbed_site_sym))

    def _create_defect_structures(
            self, defect: SimpleDefect
    ) -> Tuple[IStructure, Coords, str, Optional[IStructure],
               Optional[Tuple[PerturbedSite, ...]], Optional[str]]:

        structure = copy_to_structure(self.supercell_info.structure)

        if defect.out_atom[0] == "i":
            index = int(defect.out_atom[1:]) - 1
            site = self.supercell_info.interstitials[index]
            cutoff = self.supercell_info.interstitial_coords(index).cutoff
            coords = self.supercell_info.interstitials[index].frac_coords
        else:
            site = self.supercell_info.sites[defect.out_atom]
            cutoff = self.supercell_info.coords(defect.out_atom).cutoff
            removed_site_index = site.equivalent_atoms[0]
            coords = structure.pop(removed_site_index).frac_coords

        if defaults.displace_distance:
            p_structure, p_sites = perturb_structure(structure, coords, cutoff)

            p_site_symmetry = StructureSymmetrizer(
                p_structure,
                defaults.symmetry_length_tolerance,
                defaults.symmetry_angle_tolerance).point_group

            if defect.in_atom:
                add_atom_to_structure(structure, defect.in_atom, coords)
                add_atom_to_structure(p_structure, defect.in_atom, coords)

            return (to_istructure(structure), tuple(coords), site.site_symmetry,
                    to_istructure(p_structure), p_sites, p_site_symmetry)
        else:
            if defect.in_atom:
                add_atom_to_structure(structure, defect.in_atom, coords)
            return (to_istructure(structure), tuple(coords), site.site_symmetry,
                    None, None, None)


def copy_to_structure(structure: IStructure) -> Structure:
    return Structure.from_dict(structure.as_dict())


def to_istructure(structure: Structure) -> IStructure:
    return IStructure.from_dict(structure.as_dict())


def add_atom_to_structure(structure: Structure, elem: str, coords: List[float]):
    """In-place atom insertion to structure
    https://stackoverflow.com/questions/2361426/get-the-first-item-from-an-iterable-that-matches-a-condition
    """
    try:
        idx = next(i for i, s in enumerate(structure) if str(s.specie) == elem)
    except StopIteration:
        idx = len(structure)

    structure.insert(idx, elem, coords)


def perturb_structure(structure: Structure, center: List[float], cutoff: float
                      ) -> Tuple[Structure, Tuple[PerturbedSite, ...]]:
    """ structure perturbation

    Args:
        structure: pmg Structure class object
        center: Fractional coordinates of a central position.
        cutoff: Radius of a sphere in which atoms are perturbed.
    """
    p_structure = structure.copy()
    perturbed_sites = []
    cartesian_coords = structure.lattice.get_cartesian_coords(center)
    neighboring_atoms: List[PeriodicNeighbor] = structure.get_sites_in_sphere(
        pt=cartesian_coords, r=cutoff, include_index=True)

    for p_neighbor in neighboring_atoms:
        vector, distance = random_3d_vector(defaults.displace_distance)
        p_structure.translate_sites(p_neighbor.index, vector, frac_coords=False)
        s = PerturbedSite(
             element=str(p_neighbor.specie),
             distance=float(p_neighbor.nn_distance),
             initial_coords=tuple([float(f) for f in p_neighbor.frac_coords]),
             perturbed_coords=tuple([float(f) for f in p_structure[p_neighbor.index].frac_coords]),
             displacement=distance)
        perturbed_sites.append(s)

    return p_structure, tuple(perturbed_sites)


def random_3d_vector(max_distance: float) -> Tuple[np.ndarray, float]:
    """Random 3d vector with uniform spherical distribution with 0 <= norm <= 1.
    stackoverflow.com/questions/5408276/python-uniform-spherical-distribution
    """
    phi = np.random.uniform(0, np.pi * 2)
    cos_theta = np.random.uniform(-1, 1)
    theta = np.arccos(cos_theta)
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)
    distance = np.random.uniform(high=max_distance)
    return np.array([x, y, z]) * distance, distance



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from typing import Optional, List

import numpy as np
from numpy.linalg import det
from pydefect.defaults import defaults
from pydefect.input_maker.supercell import Supercell, TetragonalSupercells, \
    Supercells
from pydefect.input_maker.supercell_info import SupercellInfo
from pydefect.util.error_classes import NotPrimitiveError, SupercellError
from pymatgen.core import IStructure
from vise.util.centering import Centering
from vise.util.logger import get_logger
from vise.util.structure_symmetrizer import StructureSymmetrizer, Site

logger = get_logger(__name__)


class SupercellMaker:
    def __init__(self,
                 primitive_structure: IStructure,
                 matrix_to_conv_cell: Optional[List[List[int]]] = None,
                 symprec: float = defaults.symmetry_length_tolerance,
                 angle_tolerance: float = defaults.symmetry_angle_tolerance,
                 raise_error: bool = True,
                 **supercell_kwargs):

        self.primitive_structure = primitive_structure
        self.symmetrizer = StructureSymmetrizer(primitive_structure,
                                           symprec=symprec,
                                           angle_tolerance=angle_tolerance)
        if primitive_structure != self.symmetrizer.primitive:
            logger.warning(
                "The input structure is different from the primitive one,"
                "which might be due to the difference of symprec used in"
                "the pydefect and unitcell conversion.")
            logger.warning("\n".join([
                "Input lattice:",
                f"{primitive_structure.lattice}", "",
                "Primitive structure lattice:",
                f"{self.symmetrizer.primitive.lattice}", "",
                "Input structure:",
                f"{primitive_structure}", "",
                "Primitive structure:",
                f"{self.symmetrizer.primitive}"]))
            if raise_error:
                raise NotPrimitiveError

        self.sg_symbol = self.symmetrizer.spglib_sym_data["international"]
        self.conv_structure = self.symmetrizer.conventional
        crystal_system, center = str(self.symmetrizer.bravais)

        centering = Centering(center)
        self.conv_multiplicity = centering.conv_multiplicity
        self.conv_trans_mat = centering.primitive_to_conv

        self._matrix = matrix_to_conv_cell
        self._supercell_kwargs = supercell_kwargs

        self._generate_supercell(crystal_system)
        self._generate_supercell_info()

    def _generate_supercell(self, crystal_system):
        if self._matrix:
            self.supercell = Supercell(self.conv_structure, self._matrix)
        else:
            if crystal_system == "t":
                supercells = TetragonalSupercells(self.conv_structure,
                                                  **self._supercell_kwargs)
            else:
                supercells = Supercells(self.conv_structure,
                                        **self._supercell_kwargs)

            self.supercell = supercells.most_isotropic_supercell

    def _generate_supercell_info(self):
        multiplicity = int(round(det(self.transformation_matrix)))

        sites = {}
        for site_name, site in self.symmetrizer.sites.items():
            atoms = []
            for i in site.equivalent_atoms:
                atoms.extend(list(range(i*multiplicity, (i+1)*multiplicity)))
            sites[site_name] = Site(element=site.element,
                                    wyckoff_letter=site.wyckoff_letter,
                                    site_symmetry=site.site_symmetry,
                                    equivalent_atoms=atoms)

        self.supercell_info = \
            SupercellInfo(structure=self.supercell.structure,
                          space_group=self.sg_symbol,
                          transformation_matrix=self.transformation_matrix,
                          sites=sites,
                          unitcell_structure=self.primitive_structure)

    @property
    def transformation_matrix(self) -> List[List[int]]:
        """
        Definition of transformation_matrix is different from that in spglib.
        https://spglib.github.io/spglib/definition.html#transformation-matrix-boldsymbol-p-and-origin-shift-boldsymbol-p
        This is because we need to keep the consistency with the pymatgen.
        See also the unittest.
        """
        matrix = np.dot(self.supercell.matrix, self.conv_trans_mat).astype(int)
        return matrix.tolist()




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import itertools
from itertools import groupby
from typing import List

import numpy as np
import pandas as pd
from pandas import DataFrame
from pydefect.analyzer.defect_structure_info import remove_dot
from pydefect.input_maker.local_extrema import VolumetricDataLocalExtrema, \
    CoordInfo, VolumetricDataAnalyzeParams
from pydefect.util.structure_tools import Distances
from pymatgen.core import Element, Structure
from pymatgen.io.vasp import VolumetricData, Chgcar
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import squareform
from skimage.feature import peak_local_max
from vise.util.logger import get_logger
from vise.util.structure_symmetrizer import StructureSymmetrizer

logger = get_logger(__name__)


def make_local_extrema_from_volumetric_data(
        volumetric_data: VolumetricData,
        params: VolumetricDataAnalyzeParams,
        info: str = None,
        find_min: bool = True) -> VolumetricDataLocalExtrema:
    extrema = extrema_coords(volumetric_data, find_min, params)
    coord_infos = find_inequivalent_coords(volumetric_data.structure,
                                           extrema)
    return VolumetricDataLocalExtrema(volumetric_data.structure,
                                      is_min=find_min,
                                      extrema_points=coord_infos,
                                      info=info,
                                      params=params)


def find_inequivalent_coords(structure: Structure,
                             df: DataFrame) -> List[CoordInfo]:
    result = []
    initial_sg = StructureSymmetrizer(structure).sg_number
    added_structure = Structure.from_dict(structure.as_dict())
    start_index = len(structure)
    for _, column in df.iterrows():
        coords = [column.a, column.b, column.c]
        assert Element.Og not in structure.composition.elements
        added_structure.append(Element.Og, coords)
    end_index = len(added_structure)

    symmetrizer = StructureSymmetrizer(added_structure)

    if initial_sg != symmetrizer.sg_number:
        logger.warning("The symmetry has changed, meaning all the symmetry "
                       "equivalent sites do not exist.")

    sym_data = symmetrizer.spglib_sym_data
    _indices = [i for i in range(start_index, end_index)]
    repr_atom_pairs = zip(sym_data["equivalent_atoms"][_indices], _indices)

    key = lambda x: x[0]
    for _, equiv_sites in groupby(sorted(repr_atom_pairs, key=key), key=key):
        coords, quantity = [], []
        for repr_idx, atom_idx in equiv_sites:
            fcoord = added_structure[atom_idx].frac_coords
            coords.append(tuple(fcoord))
            key = "ave_value" if "ave_value" in df else "value"
            quantity.append(df[key][atom_idx - start_index])
            if repr_idx == atom_idx:
                site_sym = sym_data["site_symmetry_symbols"][atom_idx]
                distances = Distances(structure, fcoord)
                coordination = distances.coordination()

        coord_info = CoordInfo(site_symmetry=remove_dot(site_sym),
                               coordination=coordination,
                               frac_coords=coords,
                               quantities=quantity)
        result.append(coord_info)
    return result


class ChargeDensityAnalyzer:
    """
    Analyzer to find potential interstitial sites based on charge density. The
    `total` charge density is used.
    """

    def __init__(self, chgcar):
        """
        Initialization.

        Args:
            chgcar (pmg.Chgcar): input Chgcar object.
        """
        self.chgcar = chgcar
        self.structure = chgcar.structure
        self.extrema_coords = []  # list of frac_coords of local extrema
        self.extrema_type = None  # "local maxima" or "local minima"
        self._extrema_df = None  # extrema frac_coords - chg density table
        self._charge_distribution_df = None  # frac_coords - chg density table

    @classmethod
    def from_file(cls, chgcar_filename):
        """
        Init from a CHGCAR.

        :param chgcar_filename:
        :return:
        """
        chgcar = Chgcar.from_file(chgcar_filename)
        return cls(chgcar=chgcar)

    @property
    def charge_distribution_df(self):
        """
        :return: Charge distribution.
        """
        if self._charge_distribution_df is None:
            return self._get_charge_distribution_df()
        return self._charge_distribution_df

    @property
    def extrema_df(self):
        """
        :return: The extrema in charge density.
        """
        if self.extrema_type is None:
            logger.warning("Please run ChargeDensityAnalyzer.get_local_extrema first!")
        return self._extrema_df

    def _get_charge_distribution_df(self):
        """
        Return a complete table of fractional coordinates - charge density.
        """
        # Fraction coordinates and corresponding indices
        axis_grid = np.array([np.array(self.chgcar.get_axis_grid(i)) / self.structure.lattice.abc[i] for i in range(3)])
        axis_index = np.array([range(len(axis_grid[i])) for i in range(3)])

        data = {}

        for index in itertools.product(*axis_index):
            a, b, c = index
            f_coords = (axis_grid[0][a], axis_grid[1][b], axis_grid[2][c])
            data[f_coords] = self.chgcar.data["total"][a][b][c]

        # Fraction coordinates - charge density table
        df = pd.Series(data).reset_index()
        df.columns = ["a", "b", "c", "Charge Density"]
        self._charge_distribution_df = df

        return df

    def _update_extrema(self, f_coords, extrema_type, threshold_frac=None, threshold_abs=None):
        """Update _extrema_df, extrema_type and extrema_coords"""

        if threshold_frac is not None:
            if threshold_abs is not None:
                logger.warning("Filter can be either threshold_frac or threshold_abs!")  # Exit if both filter are set
                return
            if threshold_frac > 1 or threshold_frac < 0:
                raise Exception("threshold_frac range is [0, 1]!")

        # Return empty result if coords list is empty
        if len(f_coords) == 0:
            df = pd.DataFrame({}, columns=["A", "B", "C", "Chgcar"])
            self._extrema_df = df
            self.extrema_coords = []
            logger.info(f"Find {len(df)} {extrema_type}.")
            return

        data = {}
        unit = 1 / np.array(self.chgcar.dim)  # pixel along a, b, c

        for fc in f_coords:
            a, b, c = tuple(map(int, fc / unit))
            data[tuple(fc)] = self.chgcar.data["total"][a][b][c]

        df = pd.Series(data).reset_index()
        df.columns = ["a", "b", "c", "Charge Density"]
        ascending = extrema_type == "local minima"

        if threshold_abs is None:
            threshold_frac = threshold_frac if threshold_frac is not None else 1.0
            num_extrema = int(threshold_frac * len(f_coords))
            df = df.sort_values(by="Charge Density", ascending=ascending)[0:num_extrema]
            df.reset_index(drop=True, inplace=True)  # reset major index
        else:  # threshold_abs is set
            df = df.sort_values(by="Charge Density", ascending=ascending)
            df = df[df["Charge Density"] <= threshold_abs] if ascending else df[df["Charge Density"] >= threshold_abs]

        extrema_coords = []
        for row in df.iterrows():
            fc = np.array(row[1]["a":"c"])
            extrema_coords.append(fc)

        self._extrema_df = df
        self.extrema_type = extrema_type
        self.extrema_coords = extrema_coords
        logger.info(f"Find {len(df)} {extrema_type}.")

    def get_local_extrema(self, find_min=True, threshold_frac=None, threshold_abs=None):
        """
        Get all local extrema fractional coordinates in charge density,
        searching for local minimum by default. Note that sites are NOT grouped
        symmetrically.

        Args:
            find_min (bool): True to find local minimum else maximum, otherwise
                find local maximum.

            threshold_frac (float): optional fraction of extrema shown, which
                returns `threshold_frac * tot_num_extrema` extrema fractional
                coordinates based on highest/lowest intensity.

                E.g. set 0.2 to show the extrema with 20% highest or lowest
                intensity. Value range: 0 <= threshold_frac <= 1

                Note that threshold_abs and threshold_frac should not set in the
                same time.

            threshold_abs (float): optional filter. When searching for local
                minima, intensity <= threshold_abs returns; when searching for
                local maxima, intensity >= threshold_abs returns.

                Note that threshold_abs and threshold_frac should not set in the
                same time.

        Returns:
            extrema_coords (list): list of fractional coordinates corresponding
                to local extrema.
        """
        try:
            from skimage.feature import peak_local_max
        except ImportError:
            logger.warning("To find the extrema of the coordinates, "
                           "install skimage.")
            raise

        sign, extrema_type = 1, "local maxima"

        if find_min:
            sign, extrema_type = -1, "local minima"

        # Make 3x3x3 supercell
        # This is a trick to resolve the periodical boundary issue.
        total_chg = sign * self.chgcar.data["total"]
        total_chg = np.tile(total_chg, reps=(3, 3, 3))
        coordinates = peak_local_max(total_chg, min_distance=1)

        # Remove duplicated sites introduced by supercell.
        f_coords = [coord / total_chg.shape * 3 for coord in coordinates]
        f_coords = [f - 1 for f in f_coords if all(np.array(f) < 2) and all(np.array(f) >= 1)]

        # Update information
        self._update_extrema(
            f_coords,
            extrema_type,
            threshold_frac=threshold_frac,
            threshold_abs=threshold_abs,
        )

        return self.extrema_coords

    def cluster_nodes(self, tol=0.2):
        """
        Cluster nodes that are too close together using a tol.

        Args:
            tol (float): A distance tolerance. PBC is taken into account.
        """
        lattice = self.structure.lattice
        vf_coords = self.extrema_coords

        if len(vf_coords) == 0:
            if self.extrema_type is None:
                logger.warning("Please run ChargeDensityAnalyzer.get_local_extrema first!")
                return None
            new_f_coords = []
            self._update_extrema(new_f_coords, self.extrema_type)
            return new_f_coords

        # Manually generate the distance matrix (which needs to take into
        # account PBC.
        dist_matrix = np.array(lattice.get_all_distances(vf_coords, vf_coords))
        dist_matrix = (dist_matrix + dist_matrix.T) / 2

        for i in range(len(dist_matrix)):
            dist_matrix[i, i] = 0
        condensed_m = squareform(dist_matrix)
        z = linkage(condensed_m)
        cn = fcluster(z, tol, criterion="distance")
        merged_fcoords = []

        for n in set(cn):
            frac_coords = []
            for i, j in enumerate(np.where(cn == n)[0]):
                if i == 0:
                    frac_coords.append(self.extrema_coords[j])
                else:
                    f_coords = self.extrema_coords[j]
                    # We need the image to combine the frac_coords properly.
                    d, image = lattice.get_distance_and_image(frac_coords[0], f_coords)
                    frac_coords.append(f_coords + image)
            merged_fcoords.append(np.average(frac_coords, axis=0))

        merged_fcoords = [f - np.floor(f) for f in merged_fcoords]
        merged_fcoords = [f * (np.abs(f - 1) > 1e-15) for f in merged_fcoords]
        # the second line for fringe cases like
        # np.array([ 5.0000000e-01 -4.4408921e-17  5.0000000e-01])
        # where the shift to [0,1) does not work due to float precision
        self._update_extrema(merged_fcoords, extrema_type=self.extrema_type)
        logger.debug(f"{len(self.extrema_coords)} vertices after combination.")
        return None

    def remove_collisions(self, min_dist=0.5):
        """
        Remove predicted sites that are too close to existing atoms in the
        structure.

        Args:
            min_dist (float): The minimum distance (in Angstrom) that
                a predicted site needs to be from existing atoms. A min_dist
                with value <= 0 returns all sites without distance checking.
        """
        s_f_coords = self.structure.frac_coords
        f_coords = self.extrema_coords
        if len(f_coords) == 0:
            if self.extrema_type is None:
                logger.warning("Please run ChargeDensityAnalyzer.get_local_extrema first!")
                return None
            new_f_coords = []
            self._update_extrema(new_f_coords, self.extrema_type)
            return new_f_coords

        dist_matrix = self.structure.lattice.get_all_distances(f_coords, s_f_coords)
        all_dist = np.min(dist_matrix, axis=1)
        new_f_coords = []

        for i, f in enumerate(f_coords):
            if all_dist[i] > min_dist:
                new_f_coords.append(f)
        self._update_extrema(new_f_coords, self.extrema_type)

        return new_f_coords

    def get_structure_with_nodes(
            self,
            find_min=True,
            min_dist=0.5,
            tol=0.2,
            threshold_frac=None,
            threshold_abs=None,
    ):
        """
        Get the modified structure with the possible interstitial sites added.
        The species is set as a DummySpecies X.

        Args:
            find_min (bool): True to find local minimum else maximum, otherwise
                find local maximum.

            min_dist (float): The minimum distance (in Angstrom) that
                a predicted site needs to be from existing atoms. A min_dist
                with value <= 0 returns all sites without distance checking.

            tol (float): A distance tolerance of nodes clustering that sites too
                closed to other predicted sites will be merged. PBC is taken
                into account.

            threshold_frac (float): optional fraction of extrema, which returns
                `threshold_frac * tot_num_extrema` extrema fractional
                coordinates based on highest/lowest intensity.

                E.g. set 0.2 to insert DummySpecies atom at the extrema with 20%
                highest or lowest intensity.
                Value range: 0 <= threshold_frac <= 1

                Note that threshold_abs and threshold_frac should not set in the
                same time.

            threshold_abs (float): optional filter. When searching for local
                minima, intensity <= threshold_abs returns; when searching for
                local maxima, intensity >= threshold_abs returns.

                Note that threshold_abs and threshold_frac should not set in the
                same time.

        Returns:
            structure (Structure)
        """

        structure = self.structure.copy()
        self.get_local_extrema(
            find_min=find_min,
            threshold_frac=threshold_frac,
            threshold_abs=threshold_abs,
        )

        self.remove_collisions(min_dist)
        self.cluster_nodes(tol=tol)
        for fc in self.extrema_coords:
            structure.append("X", fc)

        return structure

    def sort_sites_by_integrated_chg(self, r=0.4):
        """
        Get the average charge density around each local minima in the charge density
        and store the result in _extrema_df
        Args:
            r (float): radius of sphere around each site to evaluate the average
        """

        if self.extrema_type is None:
            self.get_local_extrema()
        int_den = []
        for isite in self.extrema_coords:
            mask = self._dist_mat(isite) < r
            vol_sphere = self.chgcar.structure.volume * (mask.sum() / self.chgcar.ngridpts)
            chg_in_sphere = np.sum(self.chgcar.data["total"] * mask) / mask.size / vol_sphere
            int_den.append(chg_in_sphere)
        self._extrema_df["avg_charge_den"] = int_den
        self._extrema_df.sort_values(by=["avg_charge_den"], inplace=True)
        self._extrema_df.reset_index(drop=True, inplace=True)

    def _dist_mat(self, pos_frac):
        # return a matrix that contains the distances
        aa = np.linspace(0, 1, len(self.chgcar.get_axis_grid(0)), endpoint=False)
        bb = np.linspace(0, 1, len(self.chgcar.get_axis_grid(1)), endpoint=False)
        cc = np.linspace(0, 1, len(self.chgcar.get_axis_grid(2)), endpoint=False)
        AA, BB, CC = np.meshgrid(aa, bb, cc, indexing="ij")
        dist_from_pos = self.chgcar.structure.lattice.get_all_distances(
            fcoords1=np.vstack([AA.flatten(), BB.flatten(), CC.flatten()]).T,
            fcoords2=pos_frac,
        )
        return dist_from_pos.reshape(AA.shape)


def extrema_coords(volumetric_data: VolumetricData,
                   find_min: bool,
                   params: VolumetricDataAnalyzeParams) -> DataFrame:
    # try:
    #     from pymatgen.analysis.defects.utils import ChargeDensityAnalyzer
    # except ImportError:
    #     logger.warning("To find the extrema of the coordinates, "
    #                    "install pymatgen-analysis-defects code.")
    #     raise

    logger.info("When using this code, cite the paper related to skimage. "
                "see https://peerj.com/articles/453")

    result = ChargeDensityAnalyzer(chgcar=volumetric_data)
    result.get_local_extrema(threshold_frac=params.threshold_frac,
                             threshold_abs=params.threshold_abs,
                             find_min=find_min)
    if params.min_dist:
        # Remove sites near host atoms.
        result.remove_collisions(params.min_dist)
    if params.tol:
        try:
            # Cluster interstitials that are too close together using a tol.
            result.cluster_nodes(tol=params.tol)
        except ValueError:
            # error raised when only a single local extreme point.
            pass
    if params.radius:
        result.sort_sites_by_integrated_chg(r=params.radius)
    result = result.extrema_df
    if "avg_charge_den" in result.columns:
        result.rename(columns={'avg_charge_den': 'ave_value'}, inplace=True)
    result.rename(columns={'Charge Density': 'value'}, inplace=True)
    logger.info("\n" + result.__str__())
    return result




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

import pytest
from pydefect.analyzer.band_edge_states import EdgeInfo, BandEdgeOrbitalInfos, \
    PerfectBandEdgeState, OrbitalInfo, BandEdgeStates, BandEdgeState, \
    LocalizedOrbital
from pydefect.analyzer.defect_charge_info import DefectChargeInfo
from pydefect.analyzer.make_band_edge_states import make_band_edge_states, \
    orbital_diff, num_electron_in_cbm, num_hole_in_vbm


@pytest.fixture
def p_edge_state():
    vbm_info = EdgeInfo(band_idx=10, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=OrbitalInfo(energy=-1.0, orbitals={"Mn": [0.5, 0.6, 0.0, 0.0], "O": [0.0, 0.0, 0.0, 0.0]}, occupation=1.0, participation_ratio=0.1))
    cbm_info = EdgeInfo(band_idx=11, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=OrbitalInfo(energy=1.0, orbitals={"Mn": [0.0, 0.0, 0.0, 0.0], "O": [0.1, 0.2, 0.0, 0.0]}, occupation=0.0, participation_ratio=0.1))
    return PerfectBandEdgeState(vbm_info=vbm_info, cbm_info=cbm_info)


@pytest.fixture
def orb_infos():
    orbital_infos = [[[
        OrbitalInfo(energy=-1.1, orbitals={"Mn": [0.5, 0.6, 0.0, 0.0],
                                           "O": [0.0, 0.0, 0.0, 0.0]},
                    occupation=1.0, participation_ratio=0.1),
        OrbitalInfo(energy=-0.9, orbitals={"Mn": [0.5, 0.7, 0.0, 0.0],
                                           "O": [0.0, 0.0, 0.0, 0.0]},
                    occupation=1.0, participation_ratio=0.1),  # vbm
        OrbitalInfo(energy=0.0,  orbitals={"Mn": [0.1, 0.2, 0.0, 0.0],
                                           "O": [0.3, 0.4, 0.0, 0.0]},
                    occupation=1.0, participation_ratio=0.1),  # in-gap
        OrbitalInfo(energy=1.0,  orbitals={"Mn": [0.5, 0.8, 0.0, 0.0],
                                           "O": [0.0, 0.0, 0.0, 0.0]},
                    occupation=0.05, participation_ratio=0.1),  # in-gap
        OrbitalInfo(energy=1.2,  orbitals={"Mn": [0.0, 0.0, 0.0, 0.0],
                                           "O": [0.1, 0.3, 0.0, 0.0]},
                    occupation=0.02, participation_ratio=0.1),
        OrbitalInfo(energy=1.2,  orbitals={"Mn": [0.0, 0.0, 0.0, 0.0],
                                           "O": [0.1, 0.3, 0.0, 0.0]},
                    occupation=0.01, participation_ratio=0.1)]]]  # cbm
    return BandEdgeOrbitalInfos(kpt_coords=[(0.0, 0.0, 0.0)],
                                kpt_weights=[1.0],
                                orbital_infos=orbital_infos,
                                lowest_band_index=8,
                                fermi_level=0.5)


def test_num_electron_in_cbm(mocker):
    mock1 = mocker.Mock()
    mock2 = mocker.Mock()
    mock3 = mocker.Mock()
    mock4 = mocker.Mock()
    mock1.occupation = 0.1
    mock2.occupation = 0.2
    mock3.occupation = 0.3
    mock4.occupation = 0.4
    orb_info_by_spin = [[mock1, mock2], [mock3, mock4]]  # k-idx, band-idx
    actual = num_electron_in_cbm(orb_info_by_spin, cbm_idx=1, weights=[0.1, 0.9])
    expected = 0.1 * 0.2 + 0.9 * 0.4
    assert actual == pytest.approx(expected)

    actual = num_electron_in_cbm(orb_info_by_spin, cbm_idx=0, weights=[0.1, 0.9])
    expected = 0.1 * (0.1 + 0.2) + 0.9 * (0.3 + 0.4)
    assert actual == pytest.approx(expected)


def test_num_hole_in_vbm(mocker):
    mock1 = mocker.Mock()
    mock2 = mocker.Mock()
    mock3 = mocker.Mock()
    mock4 = mocker.Mock()
    mock1.occupation = 0.1
    mock2.occupation = 0.2
    mock3.occupation = 0.3
    mock4.occupation = 0.4
    orb_info_by_spin = [[mock1, mock2], [mock3, mock4]]  # k-idx, band-idx
    actual = num_hole_in_vbm(orb_info_by_spin, vbm_idx=0, weights=[0.1, 0.9])
    expected = 0.1 * (1 - 0.1) + 0.9 * (1 - 0.3)
    assert actual == pytest.approx(expected)

    actual = num_hole_in_vbm(orb_info_by_spin, vbm_idx=1, weights=[0.1, 0.9])
    expected = 0.1 * (2 - 0.1 - 0.2) + 0.9 * (2 - 0.3 - 0.4)
    assert actual == pytest.approx(expected)


def test_make_band_edge_state(p_edge_state, orb_infos):
    actual = make_band_edge_states(orb_infos, p_edge_state)

    vbm_info = EdgeInfo(band_idx=9, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=OrbitalInfo(
                            energy=-0.9, orbitals={"Mn": [0.5, 0.7, 0.0, 0.0],
                                                   "O": [0.0, 0.0, 0.0, 0.0]},
                            occupation=1.0, participation_ratio=0.1))
    cbm_info = EdgeInfo(band_idx=12, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=OrbitalInfo(
                            energy=1.2, orbitals={"Mn": [0.0, 0.0, 0.0, 0.0],
                                                  "O": [0.1, 0.3, 0.0, 0.0]},
                            occupation=0.02, participation_ratio=0.1))
    localized_orb_1 = LocalizedOrbital(
        band_idx=10, ave_energy=0.0, occupation=1.0,
        orbitals={"Mn": [0.1, 0.2, 0.0, 0.0], "O": [0.3, 0.4, 0.0, 0.0]},
        participation_ratio=0.1)
    localized_orb_2 = LocalizedOrbital(
        band_idx=11, ave_energy=1.0, occupation=0.05,
        orbitals={"Mn": [0.5, 0.8, 0.0, 0.0], "O": [0.0, 0.0, 0.0, 0.0]},
        participation_ratio=0.1)
    expected = BandEdgeStates(
        states=[BandEdgeState(vbm_info=vbm_info,
                              cbm_info=cbm_info,
                              vbm_orbital_diff=0.09999999999999998,
                              cbm_orbital_diff=0.09999999999999998,
                              localized_orbitals=[localized_orb_1,
                                                  localized_orb_2],
                              vbm_hole_occupation=0.0,
                              cbm_electron_occupation=0.03
                              )])
    assert actual == expected


def test_make_band_edge_state_w_defect_charge_info(p_edge_state, orb_infos, mocker):
    defect_charge_info = mocker.Mock()
    defect_charge_info.localized_orbitals.return_value = [[9]]
    actual = make_band_edge_states(orb_infos, p_edge_state, defect_charge_info)
    expected = EdgeInfo(band_idx=8, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=OrbitalInfo(
                            energy=-1.1, orbitals={"Mn": [0.5, 0.6, 0.0, 0.0],
                                                   "O": [0.0, 0.0, 0.0, 0.0]},
                            occupation=1.0, participation_ratio=0.1))
    assert actual.states[0].vbm_info == expected


def test_orbital_diff(mocker):
    orb_1 = {"Mn": [0.1, 0.0, 0.0, 0.0]}
    orb_2 = {"Mn": [0.0, 0.1, 0.0, 0.0]}
    assert orbital_diff(orb_1, orb_2) == 0.2

    orb_1 = {"Mn": [0.1, 0.0, 0.0, 0.0]}
    orb_2 = {"O": [0.0, 0.1, 0.0, 0.0]}
    assert orbital_diff(orb_1, orb_2) == 0.2





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pydefect.cli.vasp.make_unitcell import make_unitcell_from_vasp
from pymatgen.io.vasp import Vasprun, Outcar


def test_unitcell(vasp_files):
    """
    HEAD OF MICROSCOPIC STATIC DIELECTRIC TENSOR (INDEPENDENT PARTICLE, excluding Hartree and local field effects)
 ------------------------------------------------------
           1.269877     0.000000    -0.000000
           0.000000     1.269877     0.000000
           0.000000     0.000000     1.269877
 ------------------------------------------------------

 MACROSCOPIC STATIC DIELECTRIC TENSOR (including local field effects in DFT)
 ------------------------------------------------------
           1.255879     0.000000    -0.000000
          -0.000000     1.255879     0.000000
          -0.000000     0.000000     1.255879
 ------------------------------------------------------
    """

    path = vasp_files / "unitcell_Ne_solid"
    unitcell = make_unitcell_from_vasp(
        vasprun_band=Vasprun(path / "vasprun-band.xml"),
        outcar_band=Outcar(path / "OUTCAR-band"),
        outcar_dielectric_clamped=Outcar(path / "OUTCAR-dielectric"),
        outcar_dielectric_ionic=Outcar(path / "OUTCAR-dielectric"),
        )

    assert unitcell.system == "Ne"
    assert unitcell.vbm == -10.3168
    assert unitcell.cbm == 1.2042
    assert unitcell.ele_dielectric_const[0][0] == 1.255879
    assert unitcell.ion_dielectric_const[0][0] == 0.0






# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pydefect.analyzer.band_edge_states import PerfectBandEdgeState, EdgeInfo, \
    OrbitalInfo
from pydefect.cli.vasp.make_perfect_band_edge_state import \
    make_perfect_band_edge_state_from_vasp
from pymatgen.io.vasp import Procar, Vasprun, Outcar


def test_make_perfect_band_edge_state_from_vasp(vasp_files):
    procar = Procar(vasp_files / "MgO_2x2x2_perfect" / "PROCAR")
    vasprun = Vasprun(vasp_files / "MgO_2x2x2_perfect" / "vasprun.xml")
    outcar = Outcar(vasp_files / "MgO_2x2x2_perfect" / "OUTCAR")
    actual = make_perfect_band_edge_state_from_vasp(procar, vasprun, outcar)

    vbm_info = EdgeInfo(
        band_idx=127, kpt_coord=(0.25, 0.25, 0.25),
        orbital_info=OrbitalInfo(energy=2.7746,
                                 orbitals={'Mg': [0.0, 0.0, 0.0, 0.0],
                                           'O': [0.0, 0.704, 0.0, 0.0]},
                                 occupation=1.0))
    cbm_info = EdgeInfo(
        band_idx=128, kpt_coord=(0.25, 0.25, 0.25),
        orbital_info=OrbitalInfo(energy=8.2034,
                                 orbitals={'Mg': [0.192, 0.0, 0.0, 0.0],
                                           'O': [0.224, 0.096, 0.0, 0.0]},
                                 occupation=0.0))
    expected = PerfectBandEdgeState(vbm_info=vbm_info, cbm_info=cbm_info)
    assert actual == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import re
from dataclasses import dataclass
from typing import List, Optional

from monty.json import MSONable


@dataclass(frozen=True)
class Defect(MSONable):
    name: str
    charges: tuple

    @classmethod
    def from_dict(cls, d):
        return cls(name=d["name"], charges=tuple(d["charges"]))

    @property
    def str_list(self):
        return ["_".join([self.name, str(charge)]) for charge in self.charges]

    @property
    def charge_list(self):
        return [charge for charge in self.charges]


class SimpleDefect(Defect):
    def __init__(self, in_atom, out_atom, charge_list):
        if in_atom is None:
            in_atom = "Va"
        super().__init__("_".join([in_atom, out_atom]), tuple(charge_list))

    @property
    def in_atom(self):
        result = self.name.split("_")[0]
        if result == "Va":
            return
        return result

    @property
    def out_atom(self):
        return self.name.split("_")[1]


def screen_simple_defect(defect: SimpleDefect, keywords: List[str]
                         ) -> Optional[SimpleDefect]:
    charges = []
    for charge in defect.charges:
        full_name = "_".join([defect.name, str(charge)])
        if any([re.search(keyword, full_name) for keyword in keywords]):
            charges.append(charge)
    if charges:
        return SimpleDefect(defect.in_atom, defect.out_atom, tuple(charges))
    else:
        return
